from bokeh.models import Range1d
from bokeh.models import Circle
from bokeh.models import MultiLine
from bokeh.plotting import figure
from bokeh.plotting import from_networkx
from bokeh.io import show, save
import networkx as nx


def draw_network(
    G,
    title="Timelink network",
    hover_tooltips=None,
    iterations=50,
    node_colors=None,
    width=500,
    height=500,
):
    """draws a network using Bokeh and networkx layout"""

    if hover_tooltips is None:
        hover_tooltips = [("desc", "@desc"), ("type", "@type")]
    if node_colors is None:
        node_colors = (
            "type",  # TODO this should be generic maybe bokeh has builtin
            {
                "person": "green",
                "wicky-viagem": "red",
                "jesuita-entrada": "blue",
                "*default*": "gray",
            },
        )
    # Establish which categories will appear when hovering over each node
    HOVER_TOOLTIPS = hover_tooltips

    # Color nodes
    (cattribute, node_colors) = node_colors
    for m in list(G.nodes):
        nt = G.nodes[m].get(cattribute, "*default*")
        nc = node_colors.get(nt, "gray")
        G.nodes[m]["color"] = nc

    # Create a plot — set dimensions, toolbar, and title
    plot = figure(
        tooltips=HOVER_TOOLTIPS,
        tools="pan,wheel_zoom,save,reset",
        active_scroll="wheel_zoom",
        x_range=Range1d(-20.1, 20.1),
        y_range=Range1d(-15.1, 15.1),
        plot_width=width,
        plot_height=height,
        title=title,
    )

    # Create a network graph object with spring layout
    # https://networkx.github.io/documentation/networkx-1.9/reference/generated/networkx.drawing.layout.spring_layout.html
    network_graph = from_networkx(
        G, nx.spring_layout, iterations=iterations, scale=20, center=(0, 0)
    )

    # Set node size and color
    network_graph.node_renderer.glyph = Circle(size=15, fill_color="color")

    # Set edge opacity and width
    network_graph.edge_renderer.glyph = MultiLine(line_alpha=0.5, line_width=1)

    # Add network graph to the plot
    plot.renderers.append(network_graph)
    show(plot)
    fn = save(plot, title=title, filename=f"{title}.html")
    print("timelink draw_network plot saved to file: " + fn)



import os
import pandas
import timelink
from timelink.api.database import TimelinkDatabase
from timelink.api.database import is_valid_postgres_db_name
from timelink.api.database import get_postgres_dbnames
from timelink.api.database import get_sqlite_databases
from timelink.kleio.kleio_server import KleioServer


def clean_kleiofile_df(df: pandas.DataFrame) -> pandas.DataFrame:
    # Todo: move this function to the pandas module
    # convert the column "status" to the enum value
    df["status"] = df["status"].apply(lambda x: x.value)
    df["import_status"] = df["import_status"].apply(lambda x: x.value)
    # convert the column "import_errors" to int with NA as 0
    # https://stackoverflow.com/questions/21287624/convert-pandas-column-containing-nans-to-dtype-int
    df["import_errors"] = df["import_errors"].astype("Int64")
    df["import_warnings"] = df["import_errors"].astype("Int64")
    return df.fillna(0)


class TimelinkNotebook:
    """A class to interact with the Timelink system
    from Jupyter notebooks

    Example:

    .. code-block:: python

        from timelink.notebooks import TimelinkNotebook

        tln = TimelinkNotebook()
        tln.print_info()
    """

    def __init__(
        self,
        project_name=None,
        project_home=None,
        db_type=None,
        db_name=None,
        kleio_image=None,
        kleio_version=None,
        kleio_token=None,
        kleio_update=False,
        postgres_image=None,
        postgres_version=None,
        sqlite_dir=None,
        stop_duplicates=True,
        **extra_args,
    ):
        """Create a TimelinkNotebook instance

        Setup of Kleio Server and Timelink
        database is done here.

        Several functions are provided to
        manage the kleio files and access the database.

        Args:
            project_name: name of the project. Defaults to the name of the parent directory
                        of the current directory.
            project_home: directory where kleio server looks for files;
                    defaults to the parent of the current directory.
            db_type: type of database ('sqlite' or 'postgres'). Defaults to 'sqlite'
            db_name: name of the database. Defaults to project name, normalized
            kleio_image: docker image for kleio server;
                            defaults to 'timelinkserver/kleio-server'
            kleio_version: version of kleio server. Defaults to 'latest'
            kleio_token: start kleio server with this token.
                            Defaults to None (create a new token)
            kleio_update: if True, update the kleio server image. Defaults to False
            postgres_image: docker image for postgres server. Defaults to 'postgres'
            postgres_version: version of postgres server. Defaults to 'latest'
            sqlite_dir: directory where sqlite databases are. Defaults to '../database/sqlite'
            stop_duplicates: if True, stop duplicates when importing files. Defaults to True
            **extra_args: extra arguments to pass to the TimelinkDatabase object

        Returns:
            A TimelinkNotebook object
        """
        self.project_name = project_name
        self.project_home = project_home
        self.db_type = db_type
        self.db_name = db_name
        self.sqlite_dir = sqlite_dir
        self.kleio_image = kleio_image
        self.kleio_version = kleio_version
        self.postgres_image = postgres_image
        self.postgres_version = postgres_version

        if self.project_home is None:
            self.project_home = KleioServer.find_local_kleio_home()
        if self.project_name is None:
            self.project_name = os.path.basename(os.path.dirname(os.getcwd()))
        if self.db_type is None:
            self.db_type = "sqlite"
        if self.db_name is None:
            self.db_name = self.project_name.replace("-", "_").replace(" ", "_")
        if self.kleio_image is None:
            self.kleio_image = "timelinkserver/kleio-server"
        if self.kleio_version is None:
            self.kleio_version = "latest"
        if self.sqlite_dir is None:
            self.sqlite_dir = os.path.join(self.project_home, "database", "sqlite")
            # create the directory if it does not exist
            if not os.path.exists(self.sqlite_dir):
                os.makedirs(self.sqlite_dir)
        if self.postgres_image is None:
            self.postgres_image = "postgres"
        if self.postgres_version is None:
            self.postgres_version = "latest"

        if self.db_type == "postgres":
            if not is_valid_postgres_db_name(self.db_name):
                raise ValueError(f"Invalid database name: {self.db_name}")

        self.db: TimelinkDatabase = TimelinkDatabase(
            db_name=self.db_name,
            db_type=self.db_type,
            db_path=self.sqlite_dir,
            kleio_home=self.project_home,
            kleio_image=self.kleio_image,
            kleio_version=self.kleio_version,
            kleio_token=None,
            kleio_update=kleio_update,
            postgres_image=self.postgres_image,
            postgres_version=self.postgres_version,
            stop_duplicates=stop_duplicates,
            **extra_args,
        )
        self.kleio_server = self.db.get_kleio_server()

    def __repr__(self):
        return (
            f"TimelinkNotebook(project_name={self.project_name}, "
            f"project_home={self.project_home}, db_type={self.db_type}, "
            f"db_name={self.db_name}, kleio_image={self.kleio_image}, "
            f"kleio_version={self.kleio_version}, "
            f"postgres_image={self.postgres_image}, "
            f"postgres_version={self.postgres_version})"
        )

    def __str__(self):
        return (
            f"TimelinkNotebook(project_name={self.project_name}, "
            f"project_home={self.project_home}, db_type={self.db_type}, "
            f"db_name={self.db_name}, kleio_image={self.kleio_image}, "
            f"kleio_version={self.kleio_version}, "
            f"postgres_image={self.postgres_image}, "
            f"postgres_version={self.postgres_version})"
        )

    def print_info(self, show_token=False, show_password=False):
        """Print information about the TimelinkNotebook object

        Args:
            show_token: if True, show the token of the kleio server
            show_password: if True, show the password of the postgres server

        """
        info_dict = self.get_info(show_token, show_password)

        for key, value in info_dict.items():
            print(f"{key}: {value}")
        if not show_token:
            print("Call print_info(show_token=True) to show the Kleio Server token")
        if not show_password:
            print("Call print_info(show_password=True) to show the Postgres password")
        print(self.__repr__())

    def get_info(self, show_token, show_password):
        info_dict = {
            "Timelink version": timelink.version,
            "Project name": self.project_name,
            "Project home": self.project_home,
            "Database type": self.db_type,
            "Database name": self.db_name,
            "Kleio image": self.kleio_image,
        }

        kserver: KleioServer = self.db.get_kleio_server()
        if kserver is not None:
            if show_token:
                info_dict["Kleio server token"] = kserver.get_token()
            else:
                info_dict["Kleio server token"] = kserver.get_token()[:5] + "..."

            info_dict.update({
                "Kleio server URL": kserver.get_url(),
                "Kleio server home": kserver.get_kleio_home(),
            })
            if kserver.container is not None:
                info_dict["Kleio server container"] = kserver.container.name
            info_dict["Kleio version requested"] = self.kleio_version
            labels = kserver.container.labels
            build = labels.get("BUILD", "")
            version = labels.get("VERSION", "")
            build_date = labels.get("BUILD_DATE", "")
            if version != "":
                info_dict["Kleio server version"] = f"{version}.{build} ({build_date})"
        if self.db_type == "sqlite":
            info_dict["SQLite directory"] = self.sqlite_dir
        elif self.db_type == "postgres":
            if show_password:
                info_dict["Postgres password"] = self.db.db_pwd
            else:
                info_dict["Postgres password"] = "..."
            info_dict.update({
                "Postgres image": self.postgres_image,
                "Postgres version": self.postgres_version,
                "Postgres user": self.db.db_user,
            })
        db_version = self.db.get_database_version()
        if db_version is None:
            db_version = 'Not versioned with Alembic'
        info_dict["Database version"] = db_version
        return info_dict

    def get_imported_files(self, data_frame=True, **kwargs):
        """Get the list of imported files in the database

        See the get_imported_files method in the TimelinkDatabase class:
        :meth:`timelink.api.database.TimelinkDatabase.get_imported_files`

        Args:
            data_frame: if True, return a pandas DataFrame; otherwise,
                        return a list of dictionaries
            **kwargs: extra arguments to pass to the get_imported_files method
        """
        ifiles = self.db.get_imported_files(**kwargs)

        if data_frame:
            if len(ifiles) == 0:
                return pandas.DataFrame()
            ifiles_json = [f.model_dump() for f in ifiles]
            ifiles_df = pandas.DataFrame(ifiles_json)
            ifiles_df["nerrors"] = ifiles_df["nerrors"].astype("Int64")
            ifiles_df["nwarnings"] = ifiles_df["nerrors"].astype("Int64")
            return ifiles_df
        else:
            return ifiles

    def update_from_sources(self, **kwargs):
        """Update the database from a list of sources

        see: :meth:`timelink.api.database.TimelinkDatabase.update_from_sources`

        """
        self.db.update_from_sources(**kwargs)

    def get_import_status(self, data_frame=True, **kwargs):
        """Get the import status of Kleio Files

        See: :meth:`timelink.api.database.TimelinkDatabase.get_import_status`

        Returns:
            A dictionary with the status of the import process
        """
        ifiles = [f.model_dump() for f in self.db.get_import_status(**kwargs)]
        if data_frame:
            if len(ifiles) == 0:
                return pandas.DataFrame()
            # create a pandas Data frame
            ifiles_df = pandas.DataFrame(ifiles)
            # convert the column "status" to the enum value
            ifiles_df["status"] = ifiles_df["status"].apply(lambda x: x.value)
            ifiles_df["import_status"] = ifiles_df["import_status"].apply(
                lambda x: x.value
            )
            # convert the column "import_errors" to int with NA as 0
            # https://stackoverflow.com/questions/21287624/convert-pandas-column-containing-nans-to-dtype-int
            ifiles_df["import_errors"] = ifiles_df["import_errors"].astype("Int64").fillna(0)
            ifiles_df["import_warnings"] = ifiles_df["import_warnings"].astype("Int64").fillna(0)

            return ifiles_df
        else:
            return ifiles

    def get_sqlite_databases(self, sqlite_dir=None, **kwargs):
        """Get the list of sqlite databases

        Args:
            sqlite_dir: directory where the sqlite databases are located
            **kwargs: extra arguments to pass to the get_sqlite_databases function

        Returns:
            A list of sqlite databases
        """
        if sqlite_dir is None:
            sqlite_dir = self.sqlite_dir
        return get_sqlite_databases(directory_path=sqlite_dir, **kwargs)

    def get_postgres_databases(self):
        """Get the list of postgres databases

        Returns:
            A list of postgres databases
        """
        return get_postgres_dbnames()

    def table_row_count_df(self):
        """Return the row count of all tables in the database"""
        tables = self.db.table_row_count()
        tables_df = pandas.DataFrame(tables, columns=["table", "count"])
        return tables_df

    def get_file_paths(self, file_spec, rows, column):
        """Get the file paths from DataFrame of from a string

        TODO: #27 add parameter to convert the paths to absolute local paths"""
        if isinstance(file_spec, pandas.DataFrame):
            if column not in file_spec.columns:
                raise Exception(f"There is no {column} in the DataFrame")
            if rows is None:
                raise Exception("The 'rows' argument must be present")
            if type(rows) is not list:
                rows = [rows]
            file_paths = file_spec.iloc[list(rows)][column].tolist()
            return file_paths
        else:
            return []

    def get_import_rpt(
        self, file_spec: pandas.DataFrame | str, rows=None, match_path=False, **kwargs
    ):
        """Show the import report for a given file specification

        Args:
            file_spec: file specification (DataFrame or string)
                       If a DataFrame, it should have the columns 'path'
                       and the arguments 'rows' must be present
            rows: if file_spec is a DataFrane, the row number to show
            match_path: if True, the path is used to retrieve the import report;
                        if false the filename is used (default).
            **kwargs: extra arguments to pass to the show_import_rpt method
                      in the TimelinkDatabase class

        """
        rpt = ""
        if match_path:
            column = "path"
        else:
            column = "name"
        if isinstance(file_spec, pandas.DataFrame):
            paths = self.get_file_paths(file_spec, rows, column)
            for file in paths:
                rpt += file + "\n"
                rpt += self.db.get_import_rpt(file, match_path=match_path, **kwargs)
                rpt += "\n\n"
        elif isinstance(file_spec, str):
            return self.db.get_import_rpt(file_spec, match_path=match_path, **kwargs)
        else:
            raise ValueError
        return rpt

    def get_translation_report(self, file_spec, rows=None):
        """Show the translation report for a given file specification

        Args:
            file_spec: file specification (DataFrame or string)
                       If a DataFrame, it should have the columns 'rpt_url'
                       and the arguments 'rows' must be present
            rows: if file_spec is a DataFrane, the row number of interest
        """
        rpt = ""
        if isinstance(file_spec, pandas.DataFrame):
            if rows is None:
                raise ValueError(
                    "The 'rows' argument must be present "
                    "if the file_spec is a DataFrame"
                )
            elif type(rows) is not list:
                rows = [rows]
            if len(rows) == 0:
                raise ValueError(
                    "The 'rows' argument must be a non-empty list, or an integer"
                )

            paths = self.get_file_paths(file_spec, rows, "rpt_url")
            for file in paths:
                rpt += self.kleio_server.get_report(file)
        elif isinstance(file_spec, str):
            return self.kleio_server.get_report(file_spec)
        else:
            raise ValueError
        return rpt

    def get_kleio_files(self, data_frame=True, **kwargs):
        """Get the list of files in the kleio server.

        Alias to :meth:`timelink.notebooks.TimelinkNotebook.get_import_status`

        but returns a subset of the columns.


        """
        result = self.get_import_status(**kwargs)
        if len(result) == 0:
            return pandas.DataFrame()
        return result[
            [
                "path",
                "name",
                "modified",
                "status",
                "translated",
                "errors",
                "warnings",
                "import_status",
                "import_errors",
                "import_warnings",
                "import_error_rpt",
                "import_warning_rpt",
                "imported",
                "rpt_url",
                "xml_url",
            ]
        ]



"""
(c) Joaquim Carvalho 2021.
MIT License, no warranties.
"""

# pylint: disable=import-error

from sqlalchemy import Column, String, ForeignKey

from timelink.mhk.models.entity import Entity
from timelink.kleio.utilities import kleio_escape, quote_long_text


class Source(Entity):
    __tablename__ = "sources"

    id = Column(String, ForeignKey("entities.id"), primary_key=True)
    the_type = Column(String(32))
    the_date = Column(String, index=True)
    loc = Column(String)
    ref = Column(String)
    kleiofile = Column(String, index=True)
    replaces = Column(String)
    obs = Column(String)

    __mapper_args__ = {"polymorphic_identity": "source"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Source(id={sr}, "
            f'the_type="{self.the_type}", '
            f'the_date="{self.the_date}", '
            f'local="{self.loc}", '
            f'ref="{self.ref}", '
            f'kleiofile="{self.kleiofile}", '
            f'replaces="{self.replaces}", '
            f'obs="{self.obs}"'
            f")"
        )

    def __str__(self):
        r = (
            f"{self.groupname}${self.id}/{self.the_date}"
            f"/type={kleio_escape(self.the_type)}"
            f"/ref={kleio_escape(self.ref)}"
            f"/loc={kleio_escape(self.loc)}"
            f'/kleiofile={kleio_escape(self.kleiofile)}"'
            f"/replaces={self.replaces}"
        )
        if self.obs is not None and len(self.obs.strip()) > 0:
            r = f"{r}/obs={quote_long_text(self.obs)}"
        return r



from timelink.kleio.groups.kgroup import KGroup
from timelink.kleio.groups.kperson import KPerson
from timelink.kleio.groups.kobject import KObject
from timelink.kleio.groups.ksource import KSource
from timelink.kleio.groups.kact import KAct
from timelink.kleio.groups.kelement import KEntityInAttribute


class KAttribute(KGroup):
    """KAttribute(type,value,[date, obs=])

    Time varying attribute of a person, object, or other

    Attributes represent time-varying information about historical entities.
    Each attribute has a type ('address','profession', 'age'), a value and
    a date associated with the attribute.

    Elements:
        type: the type of the attribute. A String
        value: the value of the attribute. A string.
        date: the date of attribute. A string in Timelink format, optional.
        obs: a note on the attribute. A string optional.

     Kleio str definition :
        part	name=attribute ;
                guaranteed=type,value ;
                also=obs,date ;
                position=type,value,date

    """

    _name = "attr"
    _guaranteed = ["type", "value"]
    _also = ["date", "obs", "entity"]  # entity is automatically set
    _position = ["type", "value", "date"]
    _pom_class_id: str = "attribute"

    def before_include(self, container_group):
        """Method called before a new group is included in this one
        through KGroup.include(KGroup).
        """
        KGroup.before_include(self, container_group)
        self["entity"] = KEntityInAttribute("entity", container_group.id.core)
        return True

    def after_include(self, group):
        pass


KPerson.allow_as_part(KAttribute)
KObject.allow_as_part(KAttribute)
KAct.allow_as_part(KAttribute)
KSource.allow_as_part(KAttribute)
KGroup._attr_class = KAttribute



# Base class, all modules must import this to share Metadata
# for 2.0 migration see https://docs.sqlalchemy.org/en/20/changelog/whatsnew_20.html#whatsnew-20-orm-declarative-typing

from typing import Any
from sqlalchemy.orm import DeclarativeBase  # pylint: disable=import-error
from sqlalchemy.orm import declared_attr  # pylint: disable=import-error


class Base(DeclarativeBase):
    """Base class, all modules must import this to share Metadata"""

    __allow_unmapped__: bool = True  # see https://docs.sqlalchemy.org/en/20/changelog/migration_20.html#migration-20-step-six
    __table_args__: Any = {"extend_existing": True}

    id: Any
    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:  # pylint: disable=no-self-argument
        return cls.__name__.lower()



"""
mapping 'authority-register' to class aregister.
class aregister super entity table aregisters
      with attributes
           id column id baseclass id coltype varchar colsize 64 colprecision 0 pkey 1
        and
           date column the_date baseclass date coltype varchar colsize 24 colprecision 0 pkey 0
        and
           user column user baseclass user coltype varchar colsize 32 colprecision 0 pkey 0
        and
           name column name baseclass name coltype varchar colsize 254 colprecision 0 pkey 0
        and
           dbase column dbase baseclass dbase coltype varchar colsize 32 colprecision 0 pkey 0
      and
         mode column replace_mode baseclass mode coltype varchar colsize 64 colprecision 0 pkey 0
        and
           obs column obs baseclass obs coltype varchar colsize 16654 colprecision 0 pkey 0 .

(c) Joaquim Carvalho 2024
MIT License, no warranties.
"""

from typing import Optional

from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import Mapped  # pylint: disable=import-error
from sqlalchemy.orm import mapped_column  # pylint: disable=import-error

from .entity import Entity
from timelink.kleio.utilities import kleio_escape, quote_long_text


class ARegister(Entity):
    """Represent an Authorithy Register"""

    __tablename__ = "aregisters"

    id: Mapped[str] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True
    )
    the_type: Mapped[Optional[str]] = mapped_column(
        String(32),
        nullable=True,
        comment="type of the register (identifcation, user-comments, vocabulary, etc.)",
    )
    the_date: Mapped[Optional[str]] = mapped_column(String, nullable=True, index=True)
    dbase: Mapped[Optional[str]] = mapped_column(
        String, nullable=True, comment="name of the exporting database"
    )
    replace_mode: Mapped[Optional[str]] = mapped_column(
        String,
        nullable=True,
        comment="backup mode means the file is ignored by default on import",
    )
    kleiofile: Mapped[Optional[str]] = mapped_column(
        String, index=True, comment="path of the kleio file"
    )
    obs: Mapped[Optional[str]] = mapped_column(
        String, nullable=True, comment="observations"
    )

    __mapper_args__ = {"polymorphic_identity": "aregister"}

    def __repr__(self):
        sr = super().__repr__()
        return str(
            f"ARegister(id={sr}, "
            f'the_type="{self.the_type}", '
            f'the_date="{self.the_date}", '
            f'dbase="{self.dbase}"',
            f'kleiofile="{self.kleiofile}", '
            f'replace_mode="{self.replace_mode}", '
            f'obs="{self.obs}"'
            f")"
        )

    def __str__(self):
        return self.to_kleio()

    def to_kleio(self, show_contained=True, self_string=None, ident="", ident_inc="  ", width=80, **kwargs) -> str:
        if self_string is None:
            r = (
                f"{self.groupname}${self.id}/{self.the_date}"
                f"/type={kleio_escape(self.the_type)}"
                f"/replace_mode={kleio_escape(self.replace_mode)}"
                f"/the_date={kleio_escape(self.the_date)}"
                f'/kleiofile={kleio_escape(self.kleiofile)}"'
            )
            if self.obs is not None and len(self.obs.strip()) > 0:
                r = f"{r}/obs={quote_long_text(self.obs.strip(), width=width)}"
        else:
            r = self_string
        r = super().to_kleio(
            self_string=r,
            ident=ident,
            ident_inc=ident_inc,
            show_contained=show_contained,
            width=width,
        )
        return r



"""Console script for timelink.

Also provides basic mhk manager functionality.

Run with  python -m timelink.cli

"""

import os
import platform
import uvicorn
import typer
import docker
from timelink.mhk.utilities import get_mhk_info, is_mhk_installed
from timelink import migrations
from timelink import version
from timelink.api.database import TimelinkDatabase, get_postgres_dbnames
from timelink.api.database import get_sqlite_databases
from timelink.api.database import get_postgres_url
from timelink.api.database import get_sqlite_url


server: uvicorn.Server = None  # uvicorn server instance

# get the current directory
# this is used to find the alembic.ini file
# and the migrations folder
# ROOT_PATH = Path(__file__).parent.parent
current_working_directory = os.getcwd()

# We use Typer https://typer.tiangolo.com
app = typer.Typer(help=f"Timelink and MHK manager {version}")
mhk_app = typer.Typer()
app.add_typer(mhk_app, name="mhk", help="MHK legacy manager")
db_app = typer.Typer()
app.add_typer(db_app, name="db", help="Database manager")


# Start server
# see alternative methods
# https://github.com/miguelgrinberg/python-socketio/issues/332#issuecomment-712928157
# https://stackoverflow.com/questions/68603658/how-to-terminate-a-uvicorn-fastapi-application-cleanly-with-workers-2-when


def cli_header():
    typer.echo("Timelink and MHK manager")
    typer.echo(f"{current_working_directory}")


@app.command("start")
def start():
    """Starts timelink with uvicorn"""
    typer.echo("Starting Timelink")
    config = uvicorn.Config("timelink.app.main:app", port=8008, reload=True)

    global server  # pylint: disable=global-statement
    server = uvicorn.Server(config)
    server.run()
    return 0


# ================================================
# DB related commands
# These are used to manage the database migrations
# ================================================


db_index = {}
db_url = {}
avoid_db_patterns = ["_users"]


def create_db_index(avoid_patterns=None):
    """Create a dictionary of databases

    Args:
        avoid_patterns (list): list of patterns to avoid in the database name

    Returns:
        dict: dictionary of databases, key is an integer,
                    value is a tuple of the form:
                    (db_type, db_name, db_url)

    """
    pgsql_dbs = get_postgres_dbnames()
    postgres_list = [
        ("postgres", db, get_postgres_url(db)) for db in sorted(pgsql_dbs)
    ]
    sqlite_list = [
        ("sqlite", os.path.basename(db), get_sqlite_url(db))
        for db in sorted(get_sqlite_databases(current_working_directory, relative_path=False))
    ]
    all_dbs = postgres_list + sqlite_list
    if avoid_patterns:
        all_dbs = [db for db in all_dbs if all([pattern not in db[1] for pattern in avoid_patterns])]

    # make tuples of the form (int, db info) from postgres_list + sqlite_list
    enumeration = enumerate(sorted(all_dbs, key=lambda x: x[1]), 1)
    db_index = {i: db for i, db in enumeration}
    return db_index


def parse_db_url(db_url):
    # check if db_url is an integer
    if type(db_url) is int or db_url.isdigit():
        # TODO: in the future user databases may need migration too, need to refactor
        db_index = create_db_index(avoid_patterns=avoid_db_patterns)  # this is to avoid users databases
        key = int(db_url)
        db_url = (db_index[key])[2]
    return db_url


@db_app.command("list")
def db_database_list_cmd():
    """List all available databases (sqlite, postgresql, etc)"""
    global db_index
    db_index = create_db_index(avoid_patterns=avoid_db_patterns)
    typer.echo("Available data bases:")
    for i, db in db_index.items():
        db_url = db[2]
        db_version = migrations.heads(db_url)
        if len(db_version) == 0:
            db_version = ''
        else:
            db_version = f" ({db_version[0][:4]})"
        typer.echo(f" {i:>6}{db[0]:>9} {db_version:>8} {db[1]}: {db_url}")  # f-string
    typer.echo("\nList of revisions:")
    revisions = migrations.get_versions()
    nr = len(revisions)
    for i, revision in enumerate(revisions, 0):
        down = revision.down_revision or "None"
        typer.echo(f" {nr-i:>3} {revision.revision[:4]} {revision.doc} <- {down[:4]}")
    return db_index


@db_app.command("current")
def db_current_cmd(
    db_url: str,
    verbose: str = "--verbose",
):
    """Display current database revision"""
    db_url = parse_db_url(db_url)

    migrations.current(db_url, verbose)


@db_app.command("upgrade")
def db_upgrade_cmd(db_url: str, revision: str = "heads"):
    """Update database to (most recent) revision

    """
    db_url = parse_db_url(db_url)
    TimelinkDatabase(db_url=db_url)
    migrations.upgrade(db_url, revision)
    typer.echo(f"Database {db_url} upgraded to revision {revision}")


@db_app.command("create")
def db_create_cmd(db_url: str = typer.Argument(..., help="Database URL")):  # noqa: B008
    """Create a new database"""
    TimelinkDatabase(db_url=db_url)
    typer.echo(f"Database {db_url} created")


@db_app.command("heads")
def db_heads(db_url: str):
    """Return the head(s) (current revision)

    see: https://alembic.sqlalchemy.org/en/latest/api/runtime.html#alembic.runtime.migration.MigrationContext.get_current_heads
    """
    db_url = parse_db_url(db_url)
    typer.echo(migrations.heads(db_url))


@db_app.command("revision")
def db_revision(db_url: str, message: str):
    """Create a new migration script"""
    db_url = parse_db_url(db_url)
    migrations.revision(db_url, message)


@db_app.command("autogenerate")
def db_autogenerate(db_url: str, message: str):
    """Create a new migration script"""
    db_url = parse_db_url(db_url)
    migrations.autogenerate(db_url, message)


@db_app.command("history")
def db_history(verbose: bool = False):
    """Show the migration history"""
    migrations.history(verbose)


@db_app.command("stamp")
def db_stamp(db_url: str, revision: str):
    """Stamp the database to a given revision. Use 'heads' to mark the database as up-to-date"""
    db_url = parse_db_url(db_url)
    migrations.stamp(db_url, revision)

# ====================
# MHK related commands
# ====================


@mhk_app.command(name="version")
def mhk_version():
    """shows MHK manager version

    Demonstrates how to access MHK installation files and
    usage of Docker API
    """
    if is_mhk_installed():
        mhk_info = get_mhk_info()

        try:
            client = docker.from_env(version="auto")
            dv = client.version()
            mhkv = f"""    Manager version:  {mhk_info.mhk_version}
            Docker version:   {dv["Version"]}
            Host OS:          {platform.system()} {platform.release()}
            User home:        {mhk_info.user_home}
            mhk-home:         {mhk_info.mhk_home}
            mhk-home init:    {mhk_info.mhk_home_init}
            mhk-home update:  {mhk_info.mhk_home_update}
            mhk use-tag:      {mhk_info.mhk_app_env.get("TAG", "*none*")}
            mhk local host:   {mhk_info.mhk_host}
            MHK URL:          http://127.0.0.1:8080/mhk
            Kleio URL:        http://127.0.0.1:8088
            Portainer URL:    http://127.0.0.1:9000"""
            typer.echo(mhkv)
        except Exception as e:
            typer.echo(f"Could not access docker: {e}")
    else:
        type.echo("Could not find a MHK instalation")
    return 0


@mhk_app.command(name="status")
def mhk_status():
    """shows docker status information"""
    client = docker.from_env()
    dinfo = client.info()
    typer.echo(
        f"""
    Containers  :{dinfo['Containers']}
       Running  :{dinfo['ContainersRunning']}
       Paused   :{dinfo['ContainersPaused']}
       Stopped  :{dinfo['ContainersStopped']}
    """
    )
    return 0


# ================
# Main
# ================


@app.callback()
def main():
    """
    This is the timelink/MHK manager on the command line
    """

    typer.echo(f"This is the timelink/MHK manager {version}")


if __name__ == "__main__":
    app()  # pragma: no cover



# flake8: noqa: E741
from enum import Enum
from typing import Optional
from pydantic import BaseModel, Field
from datetime import datetime


class translation_status_enum(str, Enum):
    V = "V"  # valid translations
    T = "T"  # need translation (source more recent than translation)
    E = "E"  # translation with errors
    W = "W"  # translation with warnings
    P = "P"  # translation being processed
    Q = "Q"  # file queued for translation


class import_status_enum(str, Enum):
    I = "I"  # imported
    E = "E"  # imported with error
    W = "W"  # imported with warnings no errors
    N = "N"  # not imported
    U = "U"  # translation updated need to reimport


class KleioFile(BaseModel):
    """Represents the information about a kleio file and its translation and import status.

    Attributes:
        path (str): The path of the file.
        name (str): The name of the file.
        size (int): The size of the file in bytes.
        directory (str): The directory containing the file.
        modified (datetime): The last modified time of the file.
        modified_iso (datetime): The last modified time of the file in ISO format.
        modified_string (str): The last modified time of the file as a string.
        qtime (datetime): The time the file was queued for translation.
        qtime_string (str): The time the file was queued for translation as a string.
        source_url (str): The URL of the source file.
        status (translation_status_enum): The status of the file:
            - V: valid translations
            - T: need translation (source more recent than translation)
            - E: translation with errors
            - W: translation with warnings
            - P: translation being processed
            - Q: file queued for translation
        translated (Optional[datetime]): The time the file was translated.
        translated_string (Optional[str]): The time the file was translated as a string.
        errors (Optional[int]): The number of errors encountered during translation.
        warnings (Optional[int]): The number of warnings encountered during translation.
        version (Optional[str]): The version of the kleio translator.
        rpt_url (Optional[str]): The URL of the report file.
        xml_url (Optional[str]): The URL of the XML file.
        import_status (Optional[import_status_enum]): The status of the file import:
            - I: imported
            - E: imported with error
            - W: imported with warnings no errors
            - N: not imported
            - U: translation updated need to reimport
        import_errors (Optional[int]): The number of errors encountered during import.
        import_warnings (Optional[int]): The number of warnings encountered during import.
        import_error_rpt (Optional[str]): Error report from import.
        import_warning_rpt (Optional[str]): Warning report from import.
        imported (Optional[datetime]): Date of import of the file.
        imported_string (Optional[str]): Date of import of the file as a string.

    Methods:
        needs_translation(): Return True if the file needs translation.
        needs_import(): Return True if the file needs import. A file needs import if it has not been imported before or if it has been translated again since the last import.

    """

    path: str = Field(..., description="The path of the file")
    name: str = Field(..., description="The name of the file")
    size: int = Field(..., description="The size of the file in bytes")
    directory: str = Field(..., description="The directory containing the file")
    modified: datetime = Field(..., description="The last modified time of the file")
    modified_iso: datetime = Field(
        ..., description="The last modified time of the file in ISO format"
    )
    modified_string: str = Field(
        ..., description="The last modified time of the file as a string"
    )
    qtime: datetime = Field(
        ..., description="The time the file was queued for translation"
    )
    qtime_string: str = Field(
        ..., description="The time the file was queued for translation as a string"
    )
    source_url: str = Field(..., description="The URL of the source file")
    """The status of the file:
        .. code-block:: python

            V = "valid translations"
            T = "need translation (source more recent than translation)"
            E = "translation with errors"
            W = "translation with warnings"
            P = "translation being processed"
            Q = "file queued for translation"
    """
    status: translation_status_enum = Field(
        ...,
        description="""The status of the file:
                            V = valid translations
                            T = need translation (source more recent than translation)
                            E = translation with errors
                            W = translation with warnings
                            P = translation being processed
                            Q = file queued for translation""",
    )
    translated: Optional[datetime] = Field(
        None, description="The time the file was translated"
    )
    translated_string: Optional[str] = Field(
        None, description="The time the file was translated as a string"
    )
    errors: Optional[int] = Field(
        None, description="The number of errors encountered during translation"
    )
    warnings: Optional[int] = Field(
        None, description="The number of warnings encountered during translation"
    )
    version: Optional[str] = Field(
        None, description="The version of the kleio translator"
    )
    rpt_url: Optional[str] = Field(None, description="The URL of the report file")
    xml_url: Optional[str] = Field(None, description="The URL of the XML file")
    import_status: Optional[import_status_enum] = Field(
        None,
        description="""The status of the file import:
                            I = imported
                            E = imported with error
                            W = imported with warnings no errors
                            N = not imported
                            U = translation updated need to reimport""",
    )
    import_errors: Optional[int] = Field(
        None, description="The number of errors encountered during import"
    )
    import_warnings: Optional[int] = Field(
        None, description="The number of warnings encountered during import"
    )
    import_error_rpt: Optional[str] = Field(
        None, description="Error report from import"
    )
    import_warning_rpt: Optional[str] = Field(
        None, description="Warning report from import"
    )
    imported: Optional[datetime] = Field(None, description="Date of import of the file")
    imported_string: Optional[str] = Field(
        None, description="Date of import of the file as a string"
    )

    def needs_translation(self):
        """Return True if the file needs translation"""
        return self.status == translation_status_enum.T

    def needs_import(self):
        """Return True if the file needs import

        A file needs import if it has not been imported before
            or if it has been translated again since the last import
        """
        if self.import_status is None:
            raise ValueError(
                "import_status is None. Call TimelinkDatabase.get_import_status() first."
            )

        return (
            self.import_status == import_status_enum.N
            or self.import_status == import_status_enum.U
        )


class ApiPermissions(str, Enum):
    """Permissions for Kleio Server operations

    """
    sources = "sources"
    kleioset = "kleioset"
    files = "files"
    structures = "structures"
    translations = "translations"
    upload = "upload"
    delete = "delete"
    mkdir = "mkdir"
    rmdir = "rmdir"


"""
        info = {
            "comment": "An user able to translate, upload and delete files, and also create and remove directories, in specific sub-directoris in kleio-home",
            "api": [
                "sources",
                "kleioset",
                "files",
                "structures",
                "translations",
                "upload",
                "delete",
                "mkdir",
                "rmdir"
                ],
            "structures": "structures/reference_sources",
            "sources": "sources/reference_sources"
        }
"""


class TokenInfo(BaseModel):
    comment: str = Field(..., description="Comment")
    api: list[ApiPermissions] = Field(..., description="API permissions")
    structures: str = Field(..., description="Structures")
    sources: str = Field(..., description="Sources")



"""Unit test package for timelink."""

# pylint: disable=unused-import, import-error
from enum import Enum
import os
from pathlib import Path
import random
import time
import warnings
import requests

import pytest

# this is used in some MHK tests
from sqlalchemy.orm import sessionmaker

from timelink.kleio.kleio_server import KleioServer
from timelink.kleio.schemas import KleioFile

# See https://docs.sqlalchemy.org/en/14
#      /orm/session_basics.html#when-do-i-make-a-sessionmaker
# Tests should import from here to access the session
#

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
KLEIO_HOME = Path(TEST_DIR, "timelink-home")

# this will be used in some MHK tests should be deprecated
Session = sessionmaker()

# conn_string local file version
sqlite_db = Path(TEST_DIR, "sqlite/tests.sqlite")

# Extract the directory path
directory = os.path.dirname(sqlite_db)

# Create the directory if it doesn't exist
if not os.path.exists(directory):
    os.makedirs(directory)

conn_string = f"sqlite:///{sqlite_db}?check_same_thread=False"

# con_string sqlite in memory version
# conn_string = "sqlite://?check_same_thread=False"

# path to reference sqlite database
# this is used to test migrations
reference_db = Path(TEST_DIR, "db/reference_db/timelink.sqlite")
reference_db_con_str = f"sqlite:///{reference_db}"

# was: not Path(Path.home(), ".mhk").is_file(), reason="MHK not present")
mhk_absent = pytest.mark.skipif(True, reason="Deprecated tests")

skip_on_travis = pytest.mark.skipif(
    os.environ.get("TRAVIS") == "true", reason="this test requires file system access for sqlite"
)


# fixture to skip test if no internet
# Usage:
# @pytest.mark.skipif(not has_internet(), reason="No internet connection available")
def has_internet():
    """Check if there is an internet connection by trying to reach GitHub's domain."""
    try:
        # Attempt to reach GitHub's server. Replace 'https://github.com' with another URL if needed.
        response = requests.head('https://github.com', timeout=5)
        return response.status_code == 200
    except requests.ConnectionError:
        # If a connection error occurs, there is no internet connection.
        return False


class KleioServerTestMode(Enum):
    LOCAL = "local"
    DOCKER = "docker"


kleio_server_mode = KleioServerTestMode.DOCKER

# determine the image of kleio-server to use.
# "kleio-server" is a local image, can be built in timelink-kleio project
#   with "make build-local". Use for debugging kleio-server fixes
#
# "timelinkserver/kleio-server" is the version in Docker Hub
#     use when testing timelink-py against last public kleio-server images
#
use_kleio_image = "kleio-server"
# use latest or specific build e.g. 12.8.586
use_kleio_version = "latest"

skip_if_local = pytest.mark.skipif(kleio_server_mode == KleioServerTestMode.LOCAL, reason="Skipping test in LOCAL mode")


def get_one_translation(kserver: KleioServer, path="", max_wait=120) -> KleioFile:
    """Get one translation from the server"""
    translations = kserver.get_translations(path=path, recurse="yes", status="V")
    if len(translations) == 0:
        need_translation = kserver.get_translations(path=path, recurse="yes", status="T")
        if len(need_translation) == 0:
            raise RuntimeError(f"No files available for translation, path={path}")

        one_translation = random.choice(need_translation)
        kserver.translate(path=one_translation.path, recurse="no")

        counter = 0
        wait_for = 1
        max_counter = max_wait / wait_for
        translations = kserver.get_translations(path=path, recurse="yes", status="V")
        in_process = kserver.get_translations(path=path, recurse="yes", status="P")
        while len(translations) == 0 and (counter < max_counter or len(in_process) > 0):  # noqa: W503
            time.sleep(wait_for)
            counter += 1
            if counter % 20 == 0:
                warnings.warn(f"Waiting for translations {one_translation.name}, counter={counter}", stacklevel=1)
            translations = kserver.get_translations(path=path, recurse="yes", status="V")
            in_process = kserver.get_translations(path=path, recurse="yes", status="P")

    if len(translations) > 0:
        kleio_file: KleioFile = random.choice(translations)
    else:
        raise RuntimeError("No translations available in server")

    return kleio_file



# Extensions to Alembic's migration environment
# To support handling of views
# JRC: Currently unused. This is the receipt for handling views in the database
#      allowing to downgrade to previous versions of views. Also applies to
#      database procedures and functions.
#      Currently we just recreate the views at each opening of database after migration
#      so we dont need to downgrade views.
# https://alembic.sqlalchemy.org/en/latest/cookbook.html#supporting-views
from alembic.operations import MigrateOperation, Operations


class ReplaceableObject:
    """A class for creating replaceable database objects.

    "We first need to create a simple object that represents the
    “CREATE View” / “DROP View” aspect of what it is we’re building.
    We can just as well put strings inside dictionaries,
    which would also work. Nevertheless, I will illustrate how
    this object helps organize the code below."

    """

    def __init__(self, name, sqltext):
        self.name = name
        self.sqltext = sqltext


class ReversibleOp(MigrateOperation):
    def __init__(self, target):
        self.target = target

    @classmethod
    def invoke_for_target(cls, operations, target):
        op = cls(target)
        return operations.invoke(op)

    def reverse(self):
        raise NotImplementedError()

    @classmethod
    def _get_object_from_version(cls, operations, ident):
        version, objname = ident.split(".")
        # operations.get_context() method locates the version of file
        # which helps in updating the view
        module = operations.get_context().script.get_revision(version).module
        obj = getattr(module, objname)
        return obj

    # The replace operation below will always run a DROP then a CREATE.
    # The arguments replaces and replace_with accept a dot-separated string,
    # which refers to a revision number and an object name, such as "abc.popular_author"
    @classmethod
    def replace(cls, operations, target, replaces=None, replace_with=None):
        if replaces:
            old_obj = cls._get_object_from_version(operations, replaces)
            drop_old = cls(old_obj).reverse()
            create_new = cls(target)
        elif replace_with:
            old_obj = cls._get_object_from_version(operations, replace_with)
            drop_old = cls(target).reverse()
            create_new = cls(old_obj)
        else:
            raise TypeError("replaces or replace_with is required")

        operations.invoke(drop_old)
        operations.invoke(create_new)


@Operations.register_operation("create_view", "invoke_for_target")
@Operations.register_operation("replace_view", "replace")
class CreateViewOp(ReversibleOp):
    def reverse(self):
        return DropViewOp(self.target)


@Operations.register_operation("drop_view", "invoke_for_target")
class DropViewOp(ReversibleOp):
    def reverse(self):
        return CreateViewOp(self.target)


@Operations.implementation_for(CreateViewOp)
def create_view(operations, operation):
    operations.execute(
        "CREATE VIEW %s AS %s" % (operation.target.name, operation.target.sqltext)
    )


@Operations.implementation_for(DropViewOp)
def drop_view(operations, operation):
    operations.execute("DROP VIEW %s" % operation.target.name)



""".. module:: utilities
   :synopsis: Various utilities for handling Kleio groups and elements

.. moduleauthor: Joaquim Ramos de Carvalho

Kleio Groups are the building blocks for transcription of historical sources.
"""

from datetime import datetime
import json
import textwrap
from os import linesep as nl
import warnings


def kleio_escape(v: str) -> str:
    """
    Checks for Kleio special characters and quotes if needed::

         >>> print(kleio_escape('normal string'))
         normal string
         >>> print(kleio_escape('oops we have a / in the middle'))
         "oops we have a / in the middle"

    """
    if v is None:
        return None
    # test if v is already quoted
    if v.strip().startswith('"'):
        return v
    s = str(v)

    if any(i in s for i in "/;=$#%\n\""):
        if '"' not in s:
            return '"' + s + '"'
        else:  # double quote in value, need triple quotes
            return '"""' + s + '"""'
    else:
        return s


def quote_long_text(txt, initial_indent=" " * 4, indent=" " * 2, width=2048, **kwargs) -> str:
    """Surround long text with triple quotes,
    wraps and indents lines if needed.
    Some of the parameters are passed on to :py:func:`textwrap.fill`.

    Sphynx style markup

    :param txt: The text to be transformed
    :type txt: str
    :param initial_indent: string to ident the first line of paragraphs.
                        Default is 4 spaces. See :py:func:`textwrap.fill`.
    :type initial_indent: str
    :param indent: string to ident the wrap lines of
                    paragraphs (after the first).
                    Default is 2 spaces. See :py:func:`textwrap.fill`.
    :type indent: str
    :param width: width of line for wrapping. See :py:func:`textwrap.fill`.
    :type width: int
    :rtype: str
    """
    if txt is None:
        return None
    if width is None:
        width = 80

    if type(txt) is not str:
        txt = str(txt)

    # check if text already is triple quoted, starts with """ and end with """
    if txt.strip().startswith('"""'):
        return txt
    if len(txt) > width or len(txt.splitlines()) > 1:
        s = '"""'
        for line in txt.splitlines():
            w = textwrap.fill(line, width=width, initial_indent=initial_indent)
            s = s + textwrap.indent(w, indent) + nl
        s = s + indent + '"""'
    elif '"' in txt:
        s = '"""' + txt + '"""'
    else:
        s = kleio_escape(txt)
    return s


def get_extra_info_from_obs(obs_text: str) -> tuple[str, dict]:
    """
    Extracts the extra information from the extra_info string
    and returns a tuple with the cleaned string and a dictionary

    :param obs_text: The string with extra information (can have other text before)
    :type obs_text: str
    :rtype: tuple[str,dict]
    """
    if obs_text is None:
        return "", {}

    if "extra_info:" in obs_text:
        extra_info = obs_text.split("extra_info:")[1].strip()
        s = obs_text.split("extra_info:")[0].strip()
        if len(extra_info) > 0:
            extra_info_dict = json.loads(extra_info)
        else:
            extra_info_dict = {}
    else:
        s = obs_text
        extra_info_dict = {}

    return s, extra_info_dict


def render_with_extra_info(element_name, element_value, extra_info, **kwargs) -> str:
    """
    Renders a Kleio element with extra information

    :param element_name: The name of the element
    :type element_name: str
    :param element_value: The value of the element from the db
    :type element_value: str
    :param extra_info: The extra information dictionary
    :type extra_info: dict, or str (wiil be handled by :py:func:`get_extra_info`)
    :param kwargs: Additional parameters for :py:func:`quote_long_text`
    :rtype: str
    """

    if type(extra_info) is not dict:
        if type(extra_info) is str:
            _notused, extra_info = get_extra_info_from_obs(extra_info)
        else:
            extra_info = {}
    if type(element_value) is not str:
        element_value = quote_long_text(element_value, **kwargs)
    extras = extra_info.get(element_name, {})
    element_comment = extras.get("comment", None)
    element_original = extras.get("original", None)

    if element_comment is not None:
        element_value = f"{element_value}#{quote_long_text(element_comment, **kwargs)}"
    if element_original is not None:
        element_value = f"{element_value}%{quote_long_text(element_original, **kwargs)}"
    return element_value


def convert_timelink_date(tl_date: str, format="%Y%m%d") -> datetime:
    """Convert a Timelink date in the format YYYYMMDD to a Python datetime

    Args:
    tl_date: a string representing a date in the format YYYYMMDD
    format: the format used to scan the date string
    Dates can be incomplete: 1586 or 1586-03 or 158603

    Zeros can be used to register unknown values: 15860000 or 15860300

    Missing information is handled in the following way:
    * If the year is missing returns None
    * If the month is missing, returns the 2nd of July of that year (middle day of the year)
    * If the day is missing, returns the 15th of the month (middle day of the month)
    """
    # return None if tl_date is None
    if tl_date is None:
        return None
    # if tl_date is not a string, return None
    if not isinstance(tl_date, str):
        return None

    # remove dashes
    tl_date_clean = tl_date.replace("-", "")
    # pad tl_date with zeros up to length 8
    tl_date_clean = tl_date_clean.ljust(8, "0")
    year, month, day = tl_date_clean[:4], tl_date_clean[4:6], tl_date_clean[6:]
    # handle the case where month is zero by setting the month to 07 and day to 02 (middle day of year)
    if year == "0000":
        return None
    elif month == "00":
        month = "07"
        day = "02"
    elif day == "00":
        day = "15"

    new_date = year + month + day
    try:
        result = datetime.strptime(new_date, format)
    except ValueError as BadDate:
        warnings.warn(f"Invalid date: {tl_date_clean} -> {BadDate}", UserWarning, stacklevel=4)
        result = None
    return result


def format_timelink_date(tl_datet) -> str:
    """Format a timelink date YYYYMMDD and variants to nice string"""
    # return empty string if tl_datet is None
    if tl_datet is None:
        return ""
    # return empty string if tl_datet is not a string
    if not isinstance(tl_datet, str):
        return ""
    # fill with zeros
    tl_datet = tl_datet.ljust(8, "0")
    # if tl_datet is '00000000' return empty string
    if tl_datet == "00000000":
        return ""
    # if date ends in '0000' return just the first 4 characters
    if tl_datet.endswith("0000"):
        return tl_datet[:4]
    # if date ends in '00' return the first 6 characters with an hifen between 4th and 5th characters
    if tl_datet.endswith("00"):
        return tl_datet[:4] + "-" + tl_datet[4:6]
    # Otherwise convert the date
    dates = tl_datet.split(':')
    if len(dates) == 2:
        py_date1 = convert_timelink_date(dates[0])
        if py_date1 is None:
            py_date1 = ''
        py_date2 = convert_timelink_date(dates[1])
        if py_date2 is None:
            py_date2 = ''
        return (py_date1.strftime("%Y-%m-%d")
                + ":"  # noqa
                + py_date2.strftime("%Y-%m-%d"))  # noqa
    else:
        py_date = convert_timelink_date(dates[0])

        if py_date is None:
            return ""
        # return date in format YYYY-MM-DD
        return py_date.strftime("%Y-%m-%d")



import pytest

from timelink.kleio.kleio_server import KleioServer

from tests import KLEIO_HOME, KleioServerTestMode, kleio_server_mode


@pytest.fixture(scope="session")
def kleio_server():
    """setup kleio server for tests

    The kleio_server local mode does not start a new Kleio Server in docker.
    Instead it will attach to a running kleio server at a specific
    url with a specific token (setup above). This is usefull to
    have the server running in swil prolog and debug interactively.

    To run tests with a local Kleio Server outside docker:
    1) set "mode" abose to KleioServerTestMode.LOCAL
    2)Run the server in Prolog loading serverStart.pl and then:

        setenv('KLEIO_ADMIN_TOKEN','mytoken').
        setup_and_run_server(run_debug_server,[port(8089)]).

    Or run run_test_server in serverStart.pl and use
    token=mytoken and port 8088. Make sure the port
    is not in use by docker.

    Then use tpsy(predicate) in the Prolog console to debug
    requests comming from the tests in this suite.

    """
    if kleio_server_mode == KleioServerTestMode.LOCAL:
        local = True
    else:
        local = False

    if local:
        token = "mytoken"
        url = "http://localhost:8088"
        server = KleioServer.attach(url, token)
    else:
        server = KleioServer.start(kleio_home=KLEIO_HOME,
                                   kleio_version="latest",
                                   kleio_external_port=8999,
                                   kleio_debug="true")

    yield server

    if not local:
        try:
            print("Stopping kleio server", server.container.name)
            server.stop()
        except Exception as e:
            print("Error stopping server", e)
            pass



""" Test timelink Pandas integrations"""

import pytest
from pathlib import Path
import pdb  # noqa  TODO: remove when no longer needed

from sqlalchemy_utils import drop_database  # pylint: disable=unused-import. # noqa

from timelink.api.database import TimelinkDatabase
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from tests import TEST_DIR

from timelink.api.models.person import Person
from timelink.kleio.groups.kelement import KType, KValue
from timelink.kleio.groups.kls import KLs

# pytestmark = skip_on_travis

test_set = [("sqlite", "test_el_name_bug_db"), ("postgres", "test_el_name_bug_db")]


@pytest.fixture(scope="module")
def dbsystem(request, kleio_server):
    """Create a database for testing"""
    db_type, db_name = request.param
    db_path = Path(TEST_DIR, "sqlite")

    database = TimelinkDatabase(db_name, db_type, db_path=db_path, echo=False)
    # attach a kleio server
    database.set_kleio_server(kleio_server)

    try:
        yield database
    finally:
        with database.session() as session:
            pass
            database.drop_db(session)
            session.close()
        if ":memory:" not in database.db_url:
            pass
            # drop_database(database.db_url)


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_element_name_bug(dbsystem):
    """Test element name bug"""
    database = dbsystem

    database.update_from_sources("projects/test-project/sources/reference_sources/issues/element_name", force=True)
    with dbsystem.session() as session:
        id = "el-antonio-de-abreu"
        ent = session.query(Person).filter(Person.id == id).one()
        for al in ent.attributes:
            print(al.to_kleio(), al.extra_info)
        attr_name_list = ent.attributes[0].get_element_names()
        print(attr_name_list)


    KTipo = KType.extend('tipo')  # noqa
    KValor = KValue.extend('valor')   # noqa
    ls = KLs.extend("ls", position=["tipo", "valor", "date"], also=["obs", "entity"])
    _als = ls("my type", "my value", "2023-01-01")
    print(_als)
    database.update_from_sources("sources/reference_sources/issues/element_name", force=True)

    with dbsystem.session() as session:
        id = "el-antonio-de-abreu"
        ent2 = session.query(Person).filter(Person.id == id).one()
        for al in ent2.attributes:
            print(al.to_kleio(), al.extra_info)
        attr2_name_list = ent.attributes[0].get_element_names()

    assert attr_name_list == attr2_name_list



"""Timelink Python package.

Timelink, formerly known as MHK (Micro History with Kleio)
is an information system
designed for processing person related
information collected from historical sources.

.. moduleauthor:: Joaquim Ramos de Carvalho



"""
import requests

from .api import models  # noqa: F401
from .api import database  # noqa: F401
from .api import views  # noqa: F401
from .api import schemas  # noqa: F401


__author__ = """Joaquim Ramos de Carvalho"""
__email__ = 'joaquimcarvalho@mpu.edu.mo'
__version__ = '1.1.21'

version = __version__


def get_latest_version(package_name="timelink"):
    """Get the latest version of a package from PyPI."""
    # Step 1: Fetch the package information from PyPI
    url = f'https://pypi.org/pypi/{package_name}/json'
    response = requests.get(url)

    # Step 2: Parse the JSON data
    if response.status_code == 200:
        data = response.json()
        latest_version = data['info']['version']
        return latest_version
    else:
        return None



# pylint: disable=import-error

from sqlalchemy import Column, String, ForeignKey, Index
from sqlalchemy.orm import relationship
from .entity import Entity


class Attribute(Entity):
    """represents an attribute of an entity.

    Attributes have a type, a value, a date and an optional observation."""

    __tablename__ = "attributes"

    id = Column(String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True)
    entity = Column(String, ForeignKey("entities.id"), index=True)
    the_type = Column(String, index=True)
    the_value = Column(String, index=True)
    the_date = Column(String, index=True)
    obs = Column(String)
    the_entity = relationship(
        "Entity",
        foreign_keys=[entity],
        back_populates="attributes",
    )

    __mapper_args__ = {
        "polymorphic_identity": "attribute",
        "inherit_condition": id == Entity.id,
    }

    __table_args__ = (Index("attributes_type_value", "the_type", "the_value"),)

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Attribute(id={sr}, "
            f'entity="{self.entity}", '
            f'the_type="{self.the_type}", '
            f'the_value="{self.the_value}", '
            f'the_date="{self.the_date}"", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        return self.to_kleio(show_contained=False)

    def to_kleio(
        self, ident="", ident_inc="  ", self_string=None, show_contained=False, **kwargs
    ):
        if self.groupname is None:
            myname = "attribute"
        else:
            myname = self.groupname
        r = f"{myname}${self.for_kleio('the_type')}"
        r += f"/{self.for_kleio('the_value')}/"
        r += f"{self.for_kleio('the_date')}"
        obs_el = self.for_kleio("obs", named=True, prefix="/", skip_if_empty=True)
        r = f"{r}{obs_el}"
        kleio = super().to_kleio(
            self_string=r,
            show_contained=False,
            ident=ident,
            ident_inc=ident_inc,
            **kwargs,
        )
        return kleio


Entity.attributes = relationship(
    Attribute, foreign_keys=[Attribute.entity], back_populates="the_entity"
)



# Test the transformation of a csv file into a kleio file

from datetime import datetime
import csv

from tests import TEST_DIR

from timelink.kleio.groups import KDate, KType, KValue, KReplace, KName, KSex
from timelink.kleio.groups import KKleio, KSource, KAct, KPerson
from timelink.kleio.groups.kls import KLs
from timelink.kleio.groups.katr import KAtr
from timelink.kleio.groups.kgroup import KGroup

csv_file_name = "leitura_bachareis.csv"
base_csv_filename = csv_file_name.replace('.csv', '')
csv_file_path = f"{TEST_DIR}/csv_data/{csv_file_name}"
kleio_file_path = f"{TEST_DIR}/timelink-home/projects/test-project/sources/csv_to_kleio"

# max rows to convert
max_rows = 23


def test_csv_to_kleio():

    # Extend the base Elements to get localized version
    KData = KDate.extend('data')  # noqa
    KSubstitui = KReplace.extend('substitui')  # noqa
    KTipo = KType.extend('tipo')  # noqa
    KValor = KValue.extend('valor')   # noqa
    KNome = KName.extend('nome')   # noqa
    KSexo = KSex.extend('sexo')   # noqa
    KEnd = KGroup.extend('end')   # noqa

    # Define localized sources based on Base Groups
    fonte = KSource.extend(
        "fonte", position=["id"], also=["tipo", "data", "ano", "substitui", "loc", "ref", "obs", "kleiofile"]
    )
    lista = KAct.extend(
        "lista",
        position=["id", "data"],
        guaranteed=["id"],
        also=["data", "tipo", "loc", "obs"],
    )

    bacharel = KPerson.extend("bacharel", position=["nome", "sexo"], guaranteed=["id", "nome", "sexo"], also=["mesmo_que", "obs"])
    ls = KLs.extend("ls", position=["type", "value", "date"], also=["obs", "entity"])
    atr = KAtr.extend("atr", position=["type", "value", "date"], also=["date", "obs", "entity"])

    # Part A
    # 1. read the csv values into a dict
    data = []
    with open(csv_file_path, mode="r", encoding="utf-8-sig") as csv_file:
        csv_reader = csv.DictReader(csv_file, delimiter=";")
        for row in csv_reader:
            data.append(row)

    # 2. clean up the data. In this case we split dates which have
    #    some ranges in the form yyyy-mm-dd/yyyy-mm-dd in two fields

    for row in data:
        if "/" in row["Dates"]:
            dates = row["Dates"].split("/")
            row["first_date"] = dates[0].strip()
            row["last_date"] = dates[1].strip()
        else:  # if no / we set the two fields to the same value
            row["first_date"] = row["Dates"].strip()
            row["last_date"] = row["Dates"].strip()

    batch_min_date = min(row["first_date"] for row in data if row["Dates"])
    batch_max_date = max(row["last_date"] for row in data if row["Dates"])
    todays_date = datetime.today().strftime('%Y-%m-%d')

    batch_size = 10
    list_lb = lista("temp")  # we update this before writing the file
    for i, person in enumerate(data[:max_rows]):
        if i % batch_size == 0:
            if i > 0:
                # create Kleio and Source for this batch
                kleio = KKleio("gacto2.str")
                fonte_lb = fonte(f"{base_csv_filename}_{i//batch_size}", obs=f'Lido de {csv_file_name}')
                kleio.include(fonte_lb)
                # update the current list
                list_lb["id"] = f"{base_csv_filename}_l{i//batch_size}"
                list_lb["data"] = f"{batch_min_date}:{batch_max_date}"
                fonte_lb.include(list_lb)

                # Write the previous kleio content to a file
                file_path = f"{kleio_file_path}/" + f"{base_csv_filename}" + f"_{i//batch_size}.cli"
                with open(file_path, "w", encoding="utf-8") as kleio_file:
                    kleio_file.write(kleio.to_kleio())

                # create a new list for the next batch
                list_lb = lista("temp")

        p = bacharel(id=f"letbach-{i+1}", nome=person["Name"], sexo='m')

        # Update min and max dates for the current batch
        if i % batch_size == 0:
            batch_min_date = person["first_date"]
            batch_max_date = person["last_date"]
        else:
            batch_min_date = min(batch_min_date, person["first_date"])
            batch_max_date = max(batch_max_date, person["last_date"])

        if person['first_date'] == person['last_date']:
            ls_date = person['first_date']
        else:
            ls_date = f"{person['first_date']}:{person['last_date']}"

        ls_leitura = ls("leitura-bacharel.data", person["Reference"], ls_date)
        atr_url = atr("leitura_bacharel.url", person["URL"], todays_date)
        p.include(ls_leitura)
        p.include(atr_url)

        list_lb.include(p)

    # Write the last batch of kleio content to a file
    file_path = f"{kleio_file_path}/" + f"{base_csv_filename}" + f"_{i//batch_size+1}.cli"
    kleio = KKleio("gacto2.str")
    fonte_lb = fonte(f"{base_csv_filename}_{i//batch_size+1}", obs=f'Lido de {csv_file_name}')
    kleio.include(fonte_lb)
    # update the current list
    list_lb["id"] = f"{base_csv_filename}_l{i//batch_size+1}"
    list_lb["data"] = f"{batch_min_date}:{batch_max_date}"
    fonte_lb.include(list_lb)
    with open(file_path, "w", encoding="utf-8") as kleio_file:
        kleio_file.write(kleio.to_kleio())

    assert kleio.to_kleio() is not None



"""update fauc

Revision ID: 89f407c298c8
Revises: Add hoc creation to recover databases before alembic migrations
Create Date: 2025-01-18 10:58:45.626725

"""

from typing import Sequence, Union
import warnings

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "89f407c298c8"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "aregisters",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "the_type",
            sa.String(length=32),
            nullable=True,
            comment="type of the register (identifcation, user-comments, vocabulary, etc.)",
        ),
        sa.Column("the_date", sa.String(), nullable=True),
        sa.Column(
            "dbase",
            sa.String(),
            nullable=True,
            comment="name of the exporting database",
        ),
        sa.Column(
            "replace_mode",
            sa.String(),
            nullable=True,
            comment="backup mode means the file is ignored by default on import",
        ),
        sa.Column(
            "kleiofile", sa.String(), nullable=True, comment="path of the kleio file"
        ),
        sa.Column("obs", sa.String(), nullable=True, comment="observations"),
        sa.ForeignKeyConstraint(
            ["id"],
            ["entities.id"],
            name=op.f("fk_aregisters_id_entities"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_aregisters")),
        if_not_exists=True,
    )
    op.create_index(
        op.f("ix_aregisters_kleiofile"), "aregisters", ["kleiofile"], unique=False,
        if_not_exists=True
    )
    op.create_index(
        op.f("ix_aregisters_the_date"), "aregisters", ["the_date"], unique=False,
        if_not_exists=True
    )

    # links
    op.drop_table("links", if_exists=True)
    op.create_table(
        "links",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("rid", sa.String(length=64), nullable=False),
        sa.Column("entity", sa.String(length=64), nullable=True),
        sa.Column("user", sa.String(length=64), nullable=False),
        sa.Column("rule", sa.String(length=4096), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "UNLINKED",
                "AUTOMATIC",
                "SOURCE",
                "MANUAL",
                "MHK",
                "N",
                "INVALID",
                name="linkstatus",
            ),
            nullable=False,
        ),
        sa.Column("source", sa.String(length=64), nullable=True),
        sa.Column("aregister", sa.String(length=64), nullable=True),
        sa.ForeignKeyConstraint(
            ["aregister"],
            ["aregisters.id"],
            name=op.f("fk_links_aregister_aregisters"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["entity"], ["entities.id"], name=op.f("fk_links_entity_entities"),

        ),
        sa.ForeignKeyConstraint(
            ["rid"],
            ["rentities.id"],
            name=op.f("fk_links_rid_rentities"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["source"],
            ["sources.id"],
            name=op.f("fk_links_source_sources"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_links")),
        sa.UniqueConstraint("rid", "entity", "user", name="unique_link")
    )
    op.create_index(
        op.f("ix_links_entity"), "links", ["entity"], unique=False, if_not_exists=True
    )
    op.create_index(
        op.f("ix_links_rid"), "links", ["rid"], unique=False, if_not_exists=True
    )
    op.create_index(
        op.f("ix_links_source"), "links", ["source"], unique=False, if_not_exists=True
    )

    op.drop_table("robjects", if_exists=True)
    op.drop_table("goods", if_exists=True)
    op.drop_table("rpersons", if_exists=True)
    op.drop_table("rgeoentities", if_exists=True)

    # contraints
    try:
        op.drop_constraint(None, "acts", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("acts") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_acts_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    try:
        op.drop_constraint(None, "attributes", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("attributes") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_attributes_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )
    with op.batch_alter_table("class_attributes") as batch_op:
        batch_op.alter_column(
            "colname",
            existing_type=sa.VARCHAR(length=32),
            nullable=False,
        )
        batch_op.alter_column(
            "colclass",
            existing_type=sa.VARCHAR(length=32),
            nullable=False,
        )
        batch_op.alter_column("coltype", existing_type=sa.VARCHAR(), nullable=False)
        batch_op.alter_column("colsize", existing_type=sa.INTEGER(), nullable=False)
        batch_op.alter_column(
            "colprecision", existing_type=sa.INTEGER(), nullable=False
        )
        batch_op.alter_column("pkey", existing_type=sa.INTEGER(), nullable=False)

    with op.batch_alter_table("classes") as batch_op:
        batch_op.alter_column("table_name", existing_type=sa.VARCHAR(), nullable=False)

    try:
        op.drop_constraint(None, "classes", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("classes") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_classes_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    # Column changes entities

    op.add_column("entities", sa.Column("the_source", sa.String(), nullable=True))

    with op.batch_alter_table("entities") as batch_op:
        batch_op.alter_column("class", existing_type=sa.VARCHAR(), nullable=False)

    try:
        op.drop_constraint(None, "entities", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    # Geoentities
    op.create_index(op.f("ix_geoentities_name"), "geoentities", ["name"], unique=False,
                    if_not_exists=True)

    try:
        op.drop_constraint(None, "geoentities", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("geoentities") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_geoentities_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    # Kleiofiles
    with op.batch_alter_table("kleiofiles") as batch_op:
        batch_op.alter_column("imported", existing_type=sa.DATETIME(), nullable=True)
        batch_op.alter_column(
            "imported_string",
            existing_type=sa.VARCHAR(length=255),
            nullable=True,
        )

    # objects
    try:
        op.drop_constraint(None, "objects", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("objects") as batch_op:
        batch_op.alter_column(
            "the_type", existing_type=sa.VARCHAR(length=32), nullable=False
        )
        batch_op.create_foreign_key(
            op.f("fk_objects_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    # persons
    op.create_index(
        op.f("ix_persons_name"), "persons", ["name"], unique=False, if_not_exists=True
    )
    try:
        op.drop_constraint(None, "persons", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("persons") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_persons_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    # relations
    try:
        op.drop_constraint(None, "relations", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("relations") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_relations_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    # Rentities
    try:
        op.drop_constraint(None, "rentities", type_="foreignkey")
    except Exception as e:
        print(f"Warning: updating constraints (normal in old databases): {e}")

    with op.batch_alter_table("rentities") as batch_op:
        batch_op.alter_column(
            "user", existing_type=sa.VARCHAR(length=64), nullable=False
        )
        batch_op.alter_column(
            "description",
            existing_type=sa.VARCHAR(length=128),
            type_=sa.String(length=4096),
            nullable=False,
        )
        batch_op.alter_column(
            "status",
            existing_type=sa.VARCHAR(length=64),
            type_=sa.Enum(
                "UNLINKED",
                "AUTOMATIC",
                "SOURCE",
                "MANUAL",
                "MHK",
                "N",
                "INVALID",
                name="linkstatus",
            ),
            nullable=False,
        )
        batch_op.create_foreign_key(
            op.f("fk_rentities_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )

    try:
        op.drop_column("rentities", "the_class")
    except Exception as e:
        warnings.warn(f"Error dropping column 'the_class' from 'rentities': {e}", stacklevel=2)

    # sources
    try:
        op.drop_constraint(None, "sources", type_="foreignkey")
    except Exception as e:
        warnings.warn(f"Warning: updating constraints (normal in old databases): {e}", stacklevel=2)

    with op.batch_alter_table("sources") as batch_op:
        batch_op.create_foreign_key(
            op.f("fk_sources_id_entities"),
            "entities",
            ["id"],
            ["id"],
            ondelete="CASCADE",
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    pass
    # ### end Alembic commands ###



""" Structure of the app inspired by
https://levelup.gitconnected.com/structuring-fastapi-project-using-3-tier-design-pattern-4d2e88a55757


.. code-block:: bash

    .
    └── app/
        ├── backend/            # Backend functionality and configs
        |   ├── config.py           # Configuration settings
        │   └── session.py          # Database session manager
        ├── models/             # SQLAlchemy models
        │   ├── auth.py             # Authentication models
        |   ├── base.py             # Base classes, mixins
        |   └── ...                 # Other service models
        ├── routers/            # API routes
        |   ├── auth.py             # Authentication routers
        │   └── ...                 # Other service routers
        ├── schemas/            # Pydantic models
        |   ├── auth.py
        │   └── ...
        ├── services/           # Business logic
        |   ├── auth.py             # Generate and verify tokens
        |   ├── base.py             # Base classes, mixins
        │   └── ...
        ├── static/             # static files
        ├── templates/          # Jinja2 templates
        ├── cli.py              # Command-line utilities
        ├── const.py            # Constants
        ├── exc.py              # Exception handlers
        └── main.py             # Application runner

More good ideias in: https://medium.com/@ketansomvanshi007/structuring-a-fastapi-app-an-in-depth-guide-cdec3b8f4710
Official info: https://fastapi.tiangolo.com/tutorial/bigger-applications/

 """



"""Test the import of Kleio files into the Timelink database.
Check tests.__init__.py for parameters

"""
# pylint: disable=import-error
import logging
import pytest

from sqlalchemy import Engine, MetaData, select
from sqlalchemy.orm import aliased

from tests import TEST_DIR, skip_on_travis

from timelink.api.models.attribute import Attribute
from timelink.api.models.system import KleioImportedFile
from timelink.api.models import base  # pylint: disable=unused-import. # noqa: F401
from timelink.api.models.base import Person
from timelink.api.database import (
    TimelinkDatabase,
)
from timelink.api.views import DropView, view, view_exists

# https://docs.pytest.org/en/latest/how-to/skipping.html
pytestmark = skip_on_travis
path_to_db_test_files = "projects/test-project/sources/database_tests"
db_path = f"{TEST_DIR}/sqlite/"
test_set = [("sqlite", "test_db"), ("postgres", "test_db")]


@pytest.fixture(scope="module")
def dbsystem(request, kleio_server):
    """Create a database for testing"""
    db_type, db_name = request.param
    # only used for sqlite databases

    database = TimelinkDatabase(db_name, db_type, db_path=db_path, echo=False)
    # attach a kleio server
    database.set_kleio_server(kleio_server)  # from tests.__init__.py
    database.update_from_sources(path_to_db_test_files)
    try:

        yield database
    finally:
        database.drop_db()
        database.session().close()


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_create_and_drop_view(dbsystem: TimelinkDatabase):
    logging.basicConfig()
    logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
    """ test the creation and dropping of views"""
    eng: Engine = dbsystem.engine
    metadata: MetaData = dbsystem.metadata
    person = Person.__table__
    attribute = Attribute.__table__

    views_before = dbsystem.view_names()
    print(views_before)

    view_name = 'tview1'

    if view_name in views_before:
        stmt = DropView(view_name).execute_if(callable_=view_exists)

        with dbsystem.session() as session:
            session.execute(stmt)
            session.commit()

    test_view = view(
        view_name,
        metadata,
        select(
            person.c.id.label("id"),
            person.c.name.label("name"),
            person.c.sex.label("sex"),
            attribute.c.the_type.label("the_type"),
            attribute.c.the_value.label("the_value"),
            attribute.c.the_date.label("the_date"),
            person.c.obs.label("pobs"),
            attribute.c.obs.label("aobs"),
        ).select_from(person.join(attribute, person.c.id == attribute.c.entity)),
    )
    dbsystem.describe(test_view, show=True)
    views_after_def = dbsystem.view_names()
    print(views_after_def)
    assert view_name not in views_after_def

    metadata.create_all(eng)
    views_after_create = dbsystem.view_names()
    print(views_after_create)
    assert view_name in views_after_create

    stmt = DropView(view_name).execute_if(callable_=view_exists)

    with dbsystem.session() as session:
        session.execute(stmt)
        session.commit()

    views_after_delete = dbsystem.view_names()
    print(views_after_delete)
    assert view_name not in views_after_delete


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_get_nfunction_view(dbsystem):
    """Test the get_nfunction_view method."""

    # Get the nfunction view
    nfunc_view = dbsystem.get_view("nfunctions")
    assert nfunc_view is not None, "nfunction view should not be None"

    # Query the nfunction view
    with dbsystem.session() as session:
        stmt = select(nfunc_view).where(nfunc_view.c.func == "pad")
        results = session.execute(stmt).all()

    assert len(results) > 0, "There should be one result in the nfunction view"


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_get_pairs_by_function(dbsystem):
    "Test de capacity to get pairs by function in the same act"
    nfunc = dbsystem.get_view("nfunctions")
    dbsystem.describe(nfunc, show=True)
    nfunc2 = aliased(nfunc)
    func1 = "senhorio"
    func2 = "enfiteuta"

    stmt = (
        select(
            nfunc.c.id_act,
            nfunc.c.act_type,
            nfunc.c.act_date,
            nfunc.c.id.label("id1"),
            nfunc.c.func.label("func1"),
            nfunc.c.name.label("name1"),
            nfunc2.c.id.label("id2"),
            nfunc2.c.func.label("func2"),
            nfunc2.c.name.label("name2"),
        )
        .join(nfunc2, onclause=(nfunc.c.id_act == nfunc2.c.id_act))
        .where(nfunc.c.func == func1)
        .where(nfunc2.c.func == func2)
    )
    print(stmt)
    with dbsystem.session() as session:
        results = session.execute(stmt).all()
        for r in results:
            print(r)
    assert len(results) > 0, "There should be one result in the nfunction view"


@pytest.mark.parametrize("dbsystem", test_set, indirect=True)
def test_describe(dbsystem: TimelinkDatabase):
    "Test the describe method"

    nv = dbsystem.get_view("nfunctions")
    dbsystem.describe(nv, show=True)
    dbsystem.describe("act", show=True)
    dbsystem.describe(KleioImportedFile, show=True)
    dbsystem.describe(Person, show=True)
    pass



"""Add extra_info to Entity

Revision ID: 48dd68d06c60
Revises: a6d2e17ecfb1
Create Date: 2024-10-27 18:51:35.393732

"""
from typing import Sequence, Union

from alembic import op
from sqlalchemy import Column, JSON

# revision identifiers, used by Alembic.
revision: str = '48dd68d06c60'
down_revision: Union[str, None] = 'a6d2e17ecfb1'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column('entities', Column('extra_info', JSON, nullable=True))


def downgrade() -> None:
    op.drop_column('entities', 'extra_info')



"""
Test for the timelink.mhk.utilities module

This will only run if a file .mhk is found in the current home dir
"""
import warnings
import pytest
from pathlib import Path

from timelink.mhk.utilities import (
    get_mhk_env,
    get_mhk_app_env,
    get_dbnames,
    is_mhk_installed,
    get_mhk_info,
)
from tests import mhk_absent

if not is_mhk_installed():
    pytest.skip("skipping MHK tests (MHK not present)", allow_module_level=True)


@mhk_absent
def test_get_mhk_env_exists():
    if is_mhk_installed():
        mhk_env = get_mhk_env()
        assert mhk_env is not None, "is_mhk_installed true but no mhk_env"
        assert len(mhk_env) > 0, "could not get any values from ~.mhk"
    else:
        warnings.warn("MHK not installed test skipped", stacklevel=2)


@mhk_absent
def test_is_mhk_installed():
    if Path(Path.home(), ".mhk").is_file():
        assert is_mhk_installed(), "should get a true"
    else:
        assert is_mhk_installed() is False


@mhk_absent
def test_get_mhk_env_vars_ok():
    if is_mhk_installed():
        mhk_env = get_mhk_env()
        assert mhk_env is not None, "is_mhk_installed true but no mhk_env"
        should_contain = [
            "mhk_home_dir",
            "HOST_MHK_USER_HOME",
            "HOST_MHK_HOME",
            "MYSQL_OPTS",
            "kleio_url",
            "mhk_url",
            "portainer_url",
        ]
        for v in should_contain:
            assert v in mhk_env.keys(), f'no "{v}" in  ~.mhk'
    else:
        warnings.warn("MHK not installed test skipped", stacklevel=2)


@mhk_absent
def test_get_mhk_app_env_exists():
    if is_mhk_installed():
        mhk_app_env = get_mhk_app_env()
        assert mhk_app_env is not None, "is_mhk_installed true but no mhk_app_env"
        assert (
            len(mhk_app_env) > 0
        ), "could not get any values from mhk-home/app/.env"  # noqa
    else:
        warnings.warn("MHK not installed test skipped", stacklevel=2)


@mhk_absent
def test_get_app_env_vars_ok():
    if is_mhk_installed():
        mhk_app_env = get_mhk_app_env()
        assert mhk_app_env is not None, "is_mhk_installed true but no mhk_app_env"
        should_contain = [
            "env_message",
            "MHK_TOMCAT_PORT",
            "MHK_KLEIO_PORT",
            "MHK_KLEIO_PORT",
            "PORTAINER_PORT",
            "MYSQL_ROOT_PASSWORD",
        ]
        for v in should_contain:
            assert v in mhk_app_env.keys(), f'no "{v}" in  ~.mhk'
    else:
        warnings.warn("MHK not installed test skipped", stacklevel=2)


@mhk_absent
def test_mhk_info():
    if is_mhk_installed():
        minfo = get_mhk_info()
        assert minfo is not None, "is_mhk_installed true but no mhk_app_env"
        assert minfo.mhk_version
        assert minfo.user_home
        assert minfo.mhk_app_env
        assert minfo.mhk_home
        assert minfo.mhk_host
        assert minfo.mhk_app_env
        assert minfo.mhk_home_init
        assert minfo.mhk_home_update
    else:
        warnings.warn("MHK not installed test skipped", stacklevel=2)


@mhk_absent
def test_get_dbnames():
    if is_mhk_installed():
        ns = get_dbnames()
        assert len(ns) > 0, "Could not get database names. Is Mysql running?"
    else:
        warnings.warn("MHK not installed test skipped", stacklevel=2)



"""
This module provides utility functions for working with SQLite databases.

Functions:
    get_sqlite_databases(directory_path: str, relative_path=True) -> list[str]:
        Get the SQLite databases in a specified directory.

    get_sqlite_url(db_path: str) -> str:
        Get the SQLite URL for a given database path.
"""
import os


def get_sqlite_databases(directory_path: str, relative_path=True) -> list[str]:
    """Get the sqlite databases in a directory
    Args:
        directory_path (str): directory path
    Returns:
        list[str]: list of sqlite databases
    """
    cd = os.getcwd()
    sqlite_databases = []
    for root, _dirs, files in os.walk(directory_path):
        for file_name in files:
            if file_name.endswith(".sqlite") or file_name.endswith(".db"):
                db_path = os.path.join(root, file_name)
                # path relative to cd
                if relative_path:
                    db_path = os.path.relpath(db_path, cd)
                sqlite_databases.append(db_path)
    return sqlite_databases


def get_sqlite_url(db_path: str) -> str:
    """Get the sqlite url for a database path
    Args:
        db_path (str): database path
    Returns:
        str: sqlite url
    """
    if db_path == ":memory:":
        return "sqlite:///:memory:"
    return f"sqlite:///{db_path}"



import os
import time
import random
import string

import pytest
from sqlalchemy import delete, select
from sqlalchemy.exc import IntegrityError

from timelink.api import database
from timelink.app.models.user import User, UserProperty  # noqa
from timelink.app.models.user_database import UserDatabase
from timelink.app.models.project import Project, ProjectAccess


@pytest.fixture(scope="module")
def dbsystem(request: pytest.FixtureRequest):
    """Create a database for testing"""
    db_type, db_name, db_path = request.param
    db_user = None
    db_pwd = None
    if db_type == "postgres":
        database.start_postgres_server()
        db_user = database.get_postgres_container_user()
        db_pwd = database.get_postgres_container_pwd()

    dbsystem = UserDatabase(
        db_name=db_name,
        db_type=db_type,
        db_user=db_user,
        db_pwd=db_pwd,
        db_path=db_path,
    )

    with dbsystem.session() as session:
        # delete all users and properties
        session.execute(delete(ProjectAccess))
        session.execute(delete(Project))
        session.execute(delete(UserProperty))
        session.execute(delete(User))
        session.commit()

    try:
        yield dbsystem
    finally:
        dbsystem.drop_db()
        dbsystem.session().close()


dbparam = pytest.mark.parametrize(
    "dbsystem",
    [
        # db_type, db_name, db_path
        ("sqlite", "test_users.sqlite", "tests/db"),
        # see https://doc.pytest.org/en/latest/how-to/skipping.html#skip-xfail-with-parametrize
        # Not working
        # (pytest.param("postgres", "tests_users", None)),
        ("postgres", "tests_users", None),
    ],
    indirect=True,
)


@dbparam
def test_create_user_with_properties(dbsystem):
    db = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db.session() as session:
        user = User(
            name="User One",
            email="one@xpto.com",
            nickname="One",
        )
        session.add(user)
        session.commit()

        random_value1 = "".join(
            random.choices(string.ascii_uppercase + string.digits, k=10)
        )
        random_value2 = "".join(
            random.choices(string.ascii_uppercase + string.digits, k=10)
        )

        property1 = UserProperty(name="property1", value=random_value1, user_id=user.id)
        property2 = UserProperty(name="property2", value=random_value2, user_id=user.id)

        session.add(property1)
        session.add(property2)
        session.commit()

        user = session.scalars(select(User).filter_by(email="one@xpto.com")).first()

        assert len(user.properties) == 2

        # compare disregarding microseconds
        assert user.updated.replace(microsecond=0) == user.created.replace(
            microsecond=0
        )

        time.sleep(1)

        user.nickname = "One(1)"
        session.commit()

        user = session.scalars(select(User).filter_by(email="one@xpto.com")).first()
        assert user.updated > user.created


@dbparam
def test_create_user_db(dbsystem):
    db: UserDatabase = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db:
        user = User(
            name="User 2",
            email="two@xpto.com",
            nickname="NumberTwo",
        )
        db.add_user(user)
        db.commit()
        user2 = db.get_user_by_name("User 2")
        assert user2 is not None


@dbparam
def test_set_user_property(dbsystem):
    db: UserDatabase = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db:  # noqa
        user = User(
            name="User 3",
            email="san@xpto.org",
            nickname="三",
        )
        db.add_user(user)
        db.commit()  # otherwise user.id is None
        db.set_user_property(user.id, "property1", "value1")
        db.set_user_property(user.id, "property2", "value2")
        db.commit()
        user = db.get_user_by_email("san@xpto.org")
        assert user is not None
        assert len(user.properties) == 2
        props = db.get_user_properties(user.id)
        assert len(props) == 2
        proval = db.get_user_property(user.id, "property1")
        assert proval.value == "value1"
        proval = db.get_user_property(user.id, "property2")
        assert proval.value == "value2"


@dbparam
def test_update_user(dbsystem):
    db: UserDatabase = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db:
        user = User(
            name="User4",
            email="four@xpto.com",
            nickname="Test Nickname",
        )
        db.add_user(user)
        db.commit()
        user = db.get_user_by_email("four@xpto.com")
        assert user.name == "User4"
        user.name = "Quarto Utilizador"
        db.update_user(user)
        db.commit()
        user = db.get_user_by_email("four@xpto.com")
        assert user.name == "Quarto Utilizador"


@dbparam
def test_duplicate_email(dbsystem):
    db: UserDatabase = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db.session() as session:
        user = User(
            name="User five",
            email="five@xpto.com",
            nickname="five",
        )
        session.add(user)
        session.commit()

        duplicate_user = User(
            name="User six",
            email="five@xpto.com",
            nickname="six",
        )

        with pytest.raises(
            IntegrityError
        ):  # Replace Exception with a more specific exception
            session.add(duplicate_user)
            session.commit()


@dbparam
def test_user_project_access(dbsystem):
    db: UserDatabase = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db.session() as session:
        user = db.get_user_by_nickname("PU1", session=session)
        if user is None:
            user = User(
                name="Project User one",
                email="project_user_one@xpto.com",
                nickname="PU1",
                projects=[],
            )
        session.add(user)
        session.commit()

        project = session.query(Project).filter_by(name="Project One").first()
        if project is None:
            project = Project(
                name="Project One",
                description="Project One description",
            )
            session.add(project)
            session.commit()

        access_level = "admin"
        db.set_user_project_access(user.id, project.id, access_level, session=session)
        session.commit()
        user = session.scalars(select(User).filter_by(nickname="PU1")).first()
        project_access_list = [(p.project.name, p.access_level)
                               for p in user.projects
                               if p.project.name == project.name]
        assert len(project_access_list) == 1, "Project access not set"

        pa: ProjectAccess = db.get_user_project_access(user.id, project.id, session=session)
        assert pa.access_level.value == access_level, "Project access level not set"

        projects = db.get_user_projects(user.id, session=session)
        assert len(projects) == 1, "User has no projects"


@dbparam
def test_get_project_name(dbsystem):
    db: UserDatabase = dbsystem
    if db.db_type == "postgres":
        if os.environ.get("TRAVIS") == "true":
            pytest.xfail("No postgres on Travis")
    with db.session() as session:
        result = db.get_project_by_name(name="Project One", session=session)
        if result is None:
            project = Project(
                name="Project One",
                description="Project One description",
            )
            session.add(project)
            session.commit()
        else:
            project = result

        project = db.get_project_by_name(name="Project One", session=session)
        assert project is not None, "Project not found"



"""
Test for timelink.kleio package
(c) Joaquim Carvalho 2021.
MIT License, no warranties.
"""
import logging

import pytest

from timelink.kleio.groups import (
    KElement,
    KPerson,
    KAbstraction,
    KGroup,
    KKleio,
    KSource,
    KAct,
)
from timelink.kleio.groups.katr import KAtr
from timelink.kleio.groups.kelement import KDate, KDay, KMonth, KYear, KType, KReplace
from timelink.kleio.groups.kls import KLs
from timelink.kleio.utilities import quote_long_text


@pytest.fixture(scope="module")
def kgroup_nested() -> KSource:
    """Returns a nested structure"""
    ks = KSource(
        "s1", type="test", loc="auc", date="2021-07-16", ref="alumni", obs="Nested"
    )
    ka1 = KAct(
        "a1",
        "test-act",
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="auc",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)
    ka2 = KAct(
        "a2",
        "test-act",
        date="2021-07-17",
        day=17,
        month=7,
        year=2021,
        loc="auc",
        ref="p.2",
        obs="Test Act",
    )
    ks.include(ka2)
    p1 = KPerson("Joaquim", "m", "p01")
    p1.attr("residencia", "Macau", date="2021-12-11")
    p2 = KPerson("Margarida", "f", "p02")
    p2.attr("residencia", "Trouxemil", date="2020-10-18")
    p1.rel(
        "parentesco", "marido", p2.name, p2.id, date="2006-01-4", obs="Ilha Terceira"
    )
    p3 = KPerson("Pedro", "m", "p03")
    p3.attr("residencia", "Rua Arménio RC", date="2021-10-21")
    ka1.include(p1)
    ka1.include(p2)
    ka1.include(p3)
    p4 = KPerson("Maria", "f", "p04")
    p5 = KPerson("Manuel", "m", "p05")
    p6 = KPerson("João", "m", "p06")
    ka2.include(p4)
    ka2.include(p5)
    ka2.include(p6)
    return ks


@pytest.fixture(scope="module")
def kgroup_source_dev() -> KSource:
    """
    KSource('dev-1718',
                 type='Episcopal visitation',
                 loc='AUC',
                 ref='III/D,1,4,4,55',
                 date='1721-10-10:1723-07-09',
                 replace='visita-1718',
                 obs=\"""Transcrition available, manuscript.\""")
    """
    ks = KSource(
        "dev-1718",
        type="Episcopal visitation",
        loc="AUC",
        ref="III/D,1,4,4,55",
        date="1721-10-10:1723-07-09",
        replace="visita-1718",
        obs="""Transcrition available, manuscript.""",
    )
    return ks


def test_kelement():
    e = KElement("id", "0001")
    assert e.name == "id", "Failed name setting"
    assert e.core == "0001", "Failed single core value"


def test_element_empty():
    e = KElement(name="empty", core=None, original=None, comment=None)
    k = e.to_kleio(name=False)
    assert k == ""
    k = e.to_kleio(name=True)
    assert k == ""


def test_element_invisible():
    k_invisible = KElement.extend("invisible_")
    ki = k_invisible(name="invisible", core="a")
    k = ki.to_kleio()
    assert k == ""
    k = ki.to_kleio(force_show=True)
    assert k == "invisible=a"


def test_element_class_extend():
    # extend date to have a synonym in another language
    kdata_pt = KDate.extend("data")
    data_class = kdata_pt
    data_class_name = data_class.__name__
    data_class_kname = data_class.name
    assert data_class_name == data_class_kname == "data"
    data_mro = data_class.__mro__
    assert data_mro
    all_element_types = KElement.all_subclasses()
    inherited_names = [
        ke.name for ke in data_mro if ke is not object and ke.name is not None
    ]
    assert len(all_element_types) > 0
    assert len(inherited_names) > 0
    date_class = data_class.__base__
    d = kdata_pt(core="2021-12-12")
    assert d.name == "data"
    # new KElement type should keep original class
    assert d.extends(date_class.name)
    # we use a different name for the date
    fda = kdata_pt.extend("fim_de_ano")
    nd = fda("2021-12-31")
    assert nd.name == "fim_de_ano"
    assert nd.core == "2021-12-31"
    # but still is as date
    fds_names = nd.inherited_names()
    assert "date" in fds_names
    assert nd.extends(date_class.name)


def test_element_class_extend2():
    date_class = KDate.name
    group_with_date = KGroup.extend("gdate", position="date")
    gd = group_with_date("2021-12-12")
    assert gd.date.core == "2021-12-12"
    assert gd.date.element_class == date_class


def test_kelement_subclasses():
    classes = KElement.all_subclasses()
    assert len(classes) > 0


def test_kelement_subclasses2():
    classes = KElement.all_subclasses()
    element_names = [el.name for el in classes]  # noqa: F841
    assert len(classes) > 0
    assert len(element_names) > 0


def test_kelement_extend():
    KAno: KYear = KYear.extend("ano")
    annus_horribilis = KAno(1580)
    assert annus_horribilis.core + 1 == 1581


def test_kelement_extend_2():
    KAno: KYear = KYear.extend("ano")
    assert KAno.name == "ano"


def test_kelement_extend_3():
    KAno: KYear = KYear.extend("ano")
    annus_horribilis = KAno(1580)
    assert annus_horribilis.extends("ano")
    assert annus_horribilis.extends("year")


def test_kelement_extend_4():
    KAno: KYear = KYear.extend("ano")
    assert KAno in KElement.get_classes_for("ano")
    assert KAno.name == "ano"


def test_kelement_extend_6():
    """many competing classes for element get the one that matches column"""
    KAno: KYear = KYear.extend("ano")
    KAno2 = KElement.extend("ano")
    KAno3 = KAno2.extend("ano")
    KAno4 = KAno3.extend("ano")
    gano = KGroup.extend("gano", position=["ano"])
    testing: KGroup = gano(2021)
    ano = testing.get_element_by_name_or_class(KYear.name)
    assert KAno.name == KAno4.name  # just to use the variables
    assert ano.core == 2021  # get the more specialized


def test_kdate():
    d = KDate("2021-12-10")
    assert d.name == "date"


def test_kdate2():
    d = KDate("2021-12-10")
    assert d.name == "date"


def test_kday1():
    day = KDay(24)
    assert day.name == "day"


def test_kday1b():
    day = KDay(0)
    assert day.core == 0


def test_kday2():
    day = KDay("24", comment="it is a nice day", original="twenty-four")
    assert day.name == "day"
    assert day.comment == "it is a nice day"
    assert day.original == "twenty-four"


def test_kday2b():
    day = KDay(("24", "it is a nice day", "twenty-four"))
    assert day.name == "day"
    assert day.core + 1 == 25
    assert day.comment == "it is a nice day"
    assert day.original == "twenty-four"


def test_kday3():
    with pytest.raises(ValueError):
        day = KDay("32")
        assert day.core == "32"


def test_kday4():
    with pytest.raises(ValueError):
        day = KDay("31x")
        assert day.core == "31x"


def test_kday5():
    day = KDay(None)
    assert day.core == 0


def test_month1():
    month = KMonth(("05", "it is a nice month", "May"))
    assert month.name == "month"
    assert month.core + 1 == 6
    assert month.comment == "it is a nice month"
    assert month.original == "May"


def test_month2():
    month = KMonth("12")
    assert month.core + 1 == 13


def test_month3():
    with pytest.raises(ValueError):
        month = KMonth(14)
        assert month.core == 14


def test_month4():
    month = KMonth(0)
    assert month.core == 0, "Did not accept zero value"


def test_month5():
    month = KMonth(None)
    assert month.core == 0, "Did not accept zero value"


def test_year_1():
    year = KYear(2021)
    assert year.core == 2021


def test_kelement_tuple():
    e = KElement(
        "name",
        (
            "Joaquim Carvalho",
            "Family name added in the margin",
            "Joachim Carvº",
        ),
    )
    assert e.name == "name", "Failed name setting"
    assert e.core == "Joaquim Carvalho", "Failed core  from tuple"
    assert e.original == "Joachim Carvº", "Failed original from tuple"
    assert e.comment == "Family name added in the margin", "Failed comment from tuple"


def test_kelement_dict():
    e = KElement(
        "name",
        (
            "Joaquim Carvalho",
            "Family name added in the margin",
            "Joachim Carvº",
        ),
    )
    ed = e.to_dict(name=True)
    assert ed["name"] == "name", "Failed name from dict"
    assert ed["core"] == "Joaquim Carvalho", "Failed core  from dict"
    assert ed["original"] == "Joachim Carvº", "Failed original from dict"
    assert (
        ed["comment"] == "Family name added in the margin"
    ), "Failed comment from dict"


def test_kelement_dict2():
    e = KElement(
        "name",
        (
            "Joaquim Carvalho",
            "Family name added in the margin",
            "Joachim Carvº",
        ),
    )
    ed = e.to_dict(name=False)
    assert len(ed) == 3, "Failed dictionary without name"


def test_kelement_optionals():
    e = KElement(
        "name",
        "Joaquim Carvalho",
        original="Joachim Carvº",
        comment="Family name added in the margin",
    )
    assert e.name == "name", "Failed name setting"
    assert e.core == "Joaquim Carvalho", "Failed core from optionals"
    assert e.original == "Joachim Carvº", "Failed original from optional"
    assert (
        e.comment == "Family name added in the margin"
    ), "Failed comment from optional"


def test_kelement_to_str_1():
    e = KElement("name", "Joaquim Carvalho")
    assert str(e) == "Joaquim Carvalho"


def test_kelement_to_int():
    e = KElement("line", "254")
    assert int(e) + 1 == 255


def test_kelement_to_kleio():
    e = KElement("name", "Joaquim \nCarvalho")
    assert e.to_kleio(name=False) == quote_long_text(
        e.core
    ), "str() failed with multiple line core"


def test_kelement_kleio_2():
    e = KElement("name", "Joaquim Carvalho", original="Joachim Cº")
    assert (
        e.to_kleio() == "name=Joaquim Carvalho%Joachim Cº"
    ), "str() failed with original option"


def test_kelement_kleio_3():
    e = KElement("name", "Joaquim Carvalho", comment="Family name in the margin")
    assert (
        e.to_kleio(name=True) == "name=Joaquim Carvalho#Family name in the margin"
    ), "str() failed with optional comment"


def test_kelement_is_empty():
    e = KElement("el", None, comment=None, original=None)
    assert e.is_empty(), "Did not detect empty element"


def test_kelement_is_empty_2():
    e = KElement("el", "something", comment=None, original=None)
    assert not e.is_empty(), "Did not detect not empty element"


def test_kelement_is_empty_3():
    e = KElement("el", None, comment="comment", original=None)
    assert not e.is_empty(), "Did not detect not empty element"


def test_kelement_to_kleio_2():
    e = KElement(
        "name",
        "Joaquim Carvalho",
        original="Joachim Carvº",
        comment="Carvº added in the margin",
    )
    assert (
        e.to_kleio() == "name=Joaquim Carvalho#Carvº added in the margin%Joachim Carvº"
    ), "bad kleio representation of Element"


def test_kgroup_extend():
    kfonte: KGroup = KGroup.extend(
        "fonte",
        position=["id"],
        also=["tipo", "data", "ano", "obs", "substitui"],
        synonyms=[("tipo", "type"), ("data", "date"), ("ano", "year")],
    )
    afonte = kfonte("f001", ano=2021, tipo="teste")
    assert afonte.id.core == "f001", "could not extend group and instantiate"


def test_kgroup_extend2():
    # these create localized version of original elements
    KData = KDate.extend("data")  # noqa: F841
    KAno = KYear.extend("ano")  # noqa: F841
    KTipo = KType.extend("tipo")  # noqa: F841
    KSubstitui = KReplace.extend("substitui")  # noqa: F841
    kfonte: KGroup = KGroup.extend(
        "fonte", position=["id"], also=["tipo", "data", "ano", "obs", "substitui"]
    )
    afonte = kfonte("f001", data="2021-12-09", ano=2021, tipo="teste", substitui="f001")
    assert afonte.data.extends("date")
    year_value = afonte.get_element_by_name_or_class("year")
    assert 2021 == year_value.core
    assert KData.__name__ == "data"
    assert KAno.__name__ == "ano"
    assert KTipo.__name__ == "tipo"
    assert KSubstitui.__name__ == "substitui"


def test_kroup_elements_allowed():
    my_group: KGroup = KGroup.extend(
        "mygrou", position="id", also=["nome", "genero", "telefone"]
    )
    selements = sorted(my_group.elements_allowed())
    assert selements == ["genero", "id", "nome", "telefone"], "wrong element list"


def test_kroup_allow_as_element():
    my_group: KGroup = KGroup.extend(
        "mygroup", position="id", also=["nome", "genero", "telefone"]
    )
    selements = sorted(my_group.elements_allowed())
    assert selements == ["genero", "id", "nome", "telefone"], "wrong element list"

    my_group.allow_as_element("profissao", also=True)
    selements = sorted(my_group.elements_allowed())
    assert selements == [
        "genero",
        "id",
        "nome",
        "profissao",
        "telefone",
    ], "wrong element list after allow_as_element"

    my_group.allow_as_element("idade", guaranteed=True)
    selements = sorted(my_group.elements_allowed())
    assert selements == [
        "genero",
        "id",
        "idade",
        "nome",
        "profissao",
        "telefone",
    ], "wrong element list after allow_as_element"
    assert my_group._guaranteed == ["idade"]

    my_group.allow_as_element("nome", position=0)
    assert my_group._position == ["nome", "id"]


def test_kgroup_include():
    p = KPerson("joaquim", "m", "jrc")
    with pytest.raises(TypeError):
        p.include("xpto")


def test_kgroup_subclasses():
    sc_before = KPerson.all_subclasses()
    n = KPerson.extend("n")
    m = KPerson.extend("m")
    f = KPerson.extend("f")
    p = n("joaquim", "m", "01")
    sc_new = [n, m, f]
    sc_after = KPerson.all_subclasses()
    print(f"sub_classes before. {sc_before}, after: {sc_after}")
    assert len(sc_after) == len(sc_before) + len(sc_new), f"Failed to register subclasses {len(sc_after)} ≠ {len(sc_before)} + 3 "
    assert KPerson.is_kgroup(p), "Failed is_kgroup test"


def test_allow_as_part_1():

    Kx = KGroup.extend("kx", guaranteed=["id"])
    Ky = KGroup.extend("ky", guaranteed=["id"], position=["id"])

    x = Kx(id="x001")
    y = Ky("y001")

    # print(x.to_kleio())
    # print(y.to_kleio())

    with pytest.raises(ValueError):
        x.include(y), "include should have failed"

    Kx.allow_as_part('ky')
    x.include(y)
    assert y in x.contains(), "include failed"


def test_allow_as_part_2():
    class Kx(KGroup):
        pass

    class Ky(KGroup):
        pass

    Kx._name = "kx"
    Kx._guaranteed = ["id"]

    Ky._name = "ky"
    Ky._guaranteed = ["id"]
    Ky._position = ["id"]

    x = Kx(id="x001")
    y = Ky("y001")

    Kx.allow_as_part(Ky)
    x.include(y)
    assert y in list(x.contains()), "include failed after class allowed as part"


def test_allow_as_part_3():
    Kx = KGroup.extend("kx")
    Ky = KGroup.extend("ky")
    Kx.allow_as_part(Ky)
    l1 = len(Kx._part)
    Kx.allow_as_part(Ky)
    Kx.allow_as_part(Ky)
    Kx.allow_as_part(Ky)
    l2 = len(Kx._part)

    assert l1 == l2, "allow_as_part added repeated group"


def test_allow_as_part_4():
    n: KPerson = KPerson.extend("n", position=["name", "sex"], guaranteed=["name"])
    j = n("joaquim", "m", id="jrc")
    pn = KPerson.extend("pn", position=["name"], guaranteed=["name"])
    n.allow_as_part(pn)
    j.include(pn("Arménio"))
    assert len(j.contains()) == 1, "Could not insert sub group"


def test_includes_group():
    N: KPerson = KPerson.extend("n", position=["name", "sex"], guaranteed=["name"])
    j: N = N("joaquim", "m", id="jrc")  # type: ignore
    j.attr("residencia", "macau", date="2021-12-16")
    atr1 = KAtr("residencia", "Coimbra", "2020-09-20")
    atr2 = KAtr("hobby", "caligrafia", "2020-09-20")
    j.include(atr1)
    j.include(atr2)
    j.attr("idade", "63", date="2021-08-08", obs="Taipa")
    PN = KPerson.extend("pn", position=["name"], guaranteed=["name"])
    N.allow_as_part(PN)
    j.include(PN("Arménio"))
    j.attr("profissao", "professor", date="2021-08-08", obs="Taipa")

    lpn = list(j.contains(group="pn"))
    pai = lpn[0]
    np = pai.name.core
    assert np == "Arménio", "Could not retrieve group by name"


def test_includes_by_part_order():
    n: KPerson = KPerson.extend("n", position=["name", "sex"], guaranteed=["name"])
    j = n("joaquim", "m", id="jrc")
    pn = KPerson.extend("pn", position=["name"], guaranteed=["name"])
    n.allow_as_part(pn)
    j.include(pn("Arménio"))
    j.attr("residencia", "Macau", date="2021-08-08", obs="Taipa")
    inc = list(j.contains())
    assert inc[-1].kname == "pn", "included groups not by part order"


def test_includes_no_arg():
    kleio = KKleio("gacto2.str")
    ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")
    kleio.include(ks)
    ka1 = KAct(
        "a1",  # id
        "test-act",  # type
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="auc",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)
    kdots = kleio.dots
    sources = kdots.sources
    source = sources[0]
    acts = source.acts
    act = acts[0]
    assert act.id == "a1", "Problem with dot notation"
    assert (
        kdots.sources[0].acts[0].id == "a1"
    ), "Could not use dots notation for includes"
    # pretty neat!
    assert kdots.source.s1.act.a1.type == "test-act", "group-id dot notation failed"


def test_kgroup_attr():
    p = KPerson("joaquim", "m", id="joaq")
    p.attr("location", "macau", "2021")
    attrs = list(p.contains())
    assert len(attrs) > 0, "attribute not included in KGroup"


def test_kgroup_to_kleio(kgroup_source_dev):
    s = (
        "source$dev-1718/date=1721-10-10:1723-07-09/loc=AUC/"
        'ref="III/D,1,4,4,55"/replace=visita-1718/type=Episcopal visitation'
        "/obs=Transcrition available, manuscript."
    )
    kleio = kgroup_source_dev.to_kleio()
    assert kleio == s, "Bad group representation"


def test_kgroup_to_kleio_empty_1():
    p = KPerson("joaquim", "m", "jrc", obs=None)
    assert p.to_kleio() == "person$joaquim/m/jrc"


def test_kgroup_get_item():
    p = KPerson("joaquim", "m", "jrc", obs="")
    assert p["name"].core == "joaquim"


def test_kgroup_get_item_check_el():
    p = KPerson("joaquim", "m", "jrc", obs="")
    with pytest.raises(ValueError):
        p["nome"], "get inexistent element should raise error"


def test_kgroup_get_item_no_check_el():
    p: KPerson = KPerson("joaquim", "m", "jrc", obs="")
    p._element_check = False
    KElement.extend("profissao")
    p["profissao"] = "Professor"
    assert p["profissao"], "Could not enter unchecked element"


def test_kgroup_set_item():
    p = KPerson("joaquim", "m", "jrc", obs="")
    p["name"] = ("joaquim", "aka jrc", "jota")
    assert p.name.core == "joaquim", "__setitem__ failed"
    assert p.name.comment == "aka jrc", "__setitem__ failed"
    assert p.name.original == "jota", "__setitem__ failed"


def test_kgroup_set_item_check_el():
    p = KPerson("joaquim", "m", "jrc", obs="")
    with pytest.raises(ValueError):
        p["profissao"] = "professor", "set illegal element should raise error"


def test_kgroup_get(kgroup_nested):
    kgroup_nested
    assert (
        kgroup_nested.get["includes"]["act"][0]["id"] == "a1"
    ), "Failed nested dictionary"


def test_kgroup_line_seq_level():
    kk = KKleio("gacto2.str")
    ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")
    kk.include(ks)
    ka1 = KAct(
        "a1",
        "test-act",
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="auc",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)
    line = ks.line
    assert line == 2, "Failed line update"
    act = ks.contains("act")[0]
    assert act.level == 3, " Failed level update"
    p1 = KPerson("Joaquim", "m", "p01")
    ka1.include(p1)
    person = act.contains("person")[0]
    assert person.level == 4, " Failed level update"
    assert person.order == 4, " Failed sequence update"


def test_kgroup_line_seq_level_2():
    kk = KKleio("gacto2.str")
    ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")
    kk.include(ks)
    ka1 = KAct(
        "a1",
        "test-act",
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="auc",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)


def test_kgroup_set_element_more_than_once():
    ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")
    assert ks.id.core == "s1"
    ks["id"] = "s1-changed"
    assert ks.id.core == "s1-changed"


def test_kgroup_inside():
    kk = KKleio("gacto2.str")
    ks = KSource("s1", type="test", loc="auc", ref="alumni", obs="Nested")
    kk.include(ks)
    ka1 = KAct(
        "a1",
        "test-act",
        date="2021-07-16",
        day=16,
        month=7,
        year=2021,
        loc="auc",
        ref="p.1",
        obs="Test Act",
    )
    ks.include(ka1)
    assert ks.inside is kk, "Failed inside builtin"
    assert ka1.inside is ks, "Failed inside builtin"


def test_kgroup_get_nested_dict_get(kgroup_nested):
    kgroup_nested
    act = kgroup_nested.get["act"]
    assert act["a1"]["date"] == "2021-07-16", "Failed nested dictionary"


def test_kgroup_get_nested_dict_get2(kgroup_nested):
    kgroup_nested
    assert kgroup_nested.get["acts"][0]["id"] == "a1", "Failed nested dictionary"


def test_kgroup_dots(kgroup_nested):
    kgroup_nested
    assert kgroup_nested.dots.includes.act[0].id == "a1", "Failed dots retrieval"


def test_kgroup_dots2(kgroup_nested):
    kgroup_nested
    assert kgroup_nested.dots.acts[0].id == "a1", "Failed dots retrieval"


def test_kgroup_dots3(kgroup_nested):
    kgroup_nested
    the_dots = kgroup_nested.dots
    anact: KGroup = the_dots.acts[0]
    inc = anact.includes
    name = inc.person[0].name
    assert name == "Joaquim", "Failed dots retrieval"


def test_kgroup_dots4(kgroup_nested):
    kgroup_nested
    assert (
        kgroup_nested.dots.acts[0].persons[0].name == "Joaquim"
    ), "Failed dots retrieval"


def test_kgroup_dots_id(kgroup_nested):
    kgroup_nested
    assert (
        kgroup_nested.dots.act.a1.person.p01.name == "Joaquim"
    ), "Failed dots retrieval"


def test_KPerson_to_json():
    p1 = KPerson("Joaquim", "m", "p01")
    p1.attr("residencia", "Macau", date="2021-12-11")
    p2 = KPerson("Margarida", "f", "p02")
    json_string = p1.to_json()
    assert json_string
    json_string = p2.to_json()
    assert json_string
    p1.rel(
        "parentesco", "marido", p2.name, p2.id, date="2006-01-4", obs="Ilha Terceira"
    )
    json_string = p1.to_json()
    assert json_string


def test_kgroup_to_json(kgroup_nested):
    for inner in kgroup_nested.contains():
        logging.log(logging.DEBUG, f"json for {inner.kname}")
        logging.log(logging.DEBUG, f"json {inner.to_json()}")
    json_string = kgroup_nested.to_json()

    assert json_string, "Could not produce json"


def test_kgroup_core_value(kgroup_source_dev):
    assert kgroup_source_dev.id.core == "dev-1718", "Failed to retrieve core value"


def test_kgroup_get_core(kgroup_source_dev):
    assert (
        kgroup_source_dev.get_core("id") == "dev-1718"
    ), "Failed to retrieve core value"


def test_kgroup_get_core2(kgroup_source_dev):
    assert (
        kgroup_source_dev.get_core("year", 2021) == 2021
    ), "Failed to retrieve default for core value"


def test_kgroup_2(kgroup_source_dev):
    assert kgroup_source_dev.id.core == "dev-1718", "Failed to retrieve core value"


def test_person_ls_rel():
    p = KPerson("joaquim", "m", "jrc", obs="aka JRC")
    p.include(KLs("location", "Macau", "2021-07-15"))
    assert (
        " ls$location/Macau/2021-07-15" in p.to_kleio().splitlines()
    ), "ls failed to show in person"


def test_synonyms_1():
    # We can map new elements to existing ones in group creation
    kleio = KKleio
    assert kleio is not None
    fonte: KSource = KSource.extend(
        "fonte",
        also=["tipo", "data", "ano", "substitui", "loc", "ref", "obs"],
        synonyms=[("tipo", "type")],
    )
    assert fonte is not None
    tipo = KElement.get_class_for("tipo")
    assert tipo is not None


def test_synonyms_2():
    # We can map new elements to existing ones in group creation
    kleio = KKleio
    assert kleio is not None
    with pytest.raises(ValueError):
        fonte: KSource = KSource.extend(
            "fonte",
            also=["tipo", "data", "ano", "substitui", "loc", "ref", "obs"],
            synonyms=[("tipo", "typer")],  # no KElement typer
        )
        assert fonte is not None  # this should never be reached


def test_kleio_stru():
    # We specialize the kleio groups for this purpose,
    # using definitions already in gacto2.str
    kleio = KKleio
    assert kleio is not None
    fonte = KSource.extend(
        "fonte",
        also=["tipo", "data", "ano", "substitui", "loc", "ref", "obs"],
        synonyms=[("tipo", "type"), ("data", "date"), ("ano", "year")],
    )
    lista = KAct.extend(
        "lista",
        position=["id", "dia", "mes", "ano"],
        guaranteed=["id", "ano", "mes", "dia"],
        also=["data", "tipo", "loc", "obs"],
        synonyms=[
            ("tipo", "type"),
            ("dia", "day"),
            ("mes", "month"),
            ("data", "date"),
            ("ano", "year"),
        ],
    )
    auc = KAbstraction.extend(
        "auc", position=["name", "type"], also=["level", "id"], guaranteed=["id"]
    )
    n = KPerson.extend(
        "n",
        position=["nome", "sexo"],
        guaranteed=["id", "nome", "sexo"],
        also=["mesmo_que", "obs"],
        synonyms=[("nome", "name"), ("sexo", "sex"), ("mesmo_que", "same_as")],
    )
    pai = KPerson.extend(
        "pai",
        position=["nome"],
        guaranteed=["nome"],
        also=["mesmo_que", "obs"],
        synonyms=[("nome", "name"), ("mesmo_que", "same_as")],
    )
    mae = KPerson.extend(
        "mae",
        position=["nome"],
        guaranteed=["nome"],
        also=["mesmo_que", "obs"],
        synonyms=[("nome", "name"), ("mesmo_que", "same_as")],
    )
    n.allow_as_part(pai)
    n.allow_as_part(mae)
    ls = KLs.extend(
        "ls",
        position=["type", "value"],
        also=["data", "obs", "entity"],
        synonyms=[("data", "date")],
    )
    atr = KAtr.extend(
        "atr",
        position=["type", "value"],
        also=["data", "obs", "entity"],
        synonyms=[("data", "date")],
    )

    kf = KKleio()
    f = fonte("f001", tipo="auc-tests")
    kf.include(f)
    l: KGroup = lista("l001", 11, 2, 2021, data="1537-1913", tipo="auc-list", loc="A")
    f.include(l)
    a = auc("alumni-record", "archeevo-record", id="xpto")
    l.include(a)
    j: KGroup = n("joaquim", "m", obs="em macau", id="jrc")
    j.include(ls("uc", "início", data=2021))
    j.include(atr("xauc", "dsd"))
    l.include(j)
    l2 = lista("l003", 18, 2, 2021, data="1537-1913", tipo="auc-list", loc="B")
    f.include(l2)
    m = n("manuel", "m", obs="em Berlin", id="mrc")
    m.include(ls("uc", "fim", data=2021))
    l2.include(m)
    json_string = j.to_json()
    assert j.nome.core == "joaquim"
    assert len(json_string) > 0



""" """

from typing import List
import pandas as pd
from sqlalchemy.sql import select
from timelink.api.database import TimelinkDatabase


def entities_with_attribute(
    the_type: str | List[str],
    the_value=None,
    column_name=None,
    entity_type="entity",
    show_elements=None,
    dates_in=None,
    name_like=None,
    filter_by=None,
    more_attributes=None,
    db: TimelinkDatabase = None,
    sql_echo=False,
):
    """Generate a pandas dataframe with entities with a given attribute

    Args:
        the_type    : type of attribute, can have SQL wildcards, string, or list
        the_value   : if present, limit to this value, can be SQL wildcard,
        entity_type : if present, limit to this entity type, string
        column_name : if present, use this name for the attribute column, otherwise use the_type
                      usefull when the_type is a list
        name_like   : if present, limit to this name, can have SQL wildcards
        show_elements: List of entity elements to add to the dataframe
        dates_in    : (after,before) if present only between those dates (exclusive)
        filter_by   : list of ids, limit to these entities
        more_attributes: add more attributes if available
        db          : A TimelinkDatabase object
        sql_echo    : if True echo the sql generated

    Example:
        # name, sex and function of people living in the same place

        neighbors = entities_with_attribute(
                                entity_type="person",
                                show_elements=["groupname","names","sex"],
                                the_type='residencia',
                                the_value="soure"
                                column_name="local", # use this istead of "residencia"
                                more_attributes=["profissao"]
                                )

    Notes:
        - The function will return a dataframe with the columns:
            id, group

    Ideas:
        Add :
            the_value_in: (list of values)
            the_value_between_inc (min, max, get >=min and <= max)
            the_value_between_exc (min, max, get >min and < max)


    """
    # We try to use an existing connection and table introspection
    # to avoid extra parameters and going to database too much
    dbsystem: TimelinkDatabase = db
    if dbsystem is None:
        raise (
            Exception(
                "db: TimelinkDatabase required. Must create  to set up a database connection"
            )
        )

    # if we dont have a name for the column we use the attribute type sanitized
    if column_name is None:
        # if the_type is a list or a string
        if type(the_type) is list:
            column_name = "_".join(the_type)
        else:
            column_name = str(the_type)
    date_column_name = f"{column_name}.date"
    obs_column_name = f"{column_name}.obs"
    type_column_name = f"{column_name}.type"
    line_column_name = f"{column_name}.line"
    level_column_name = f"{column_name}.level"
    atr_id_column_name = f"{column_name}.attr_id"
    attribute_extra_info_column_name = f"{column_name}.extra_info"

    entity_types = db.get_models_ids()
    if entity_type not in entity_types:
        raise ValueError(f"entity_type must be one of {entity_types}")

    entity_model = db.get_model(entity_type)
    # get the columns of the entity table, check if more_info is valid
    entity_columns = select(entity_model).selected_columns.keys()
    entity_id_col = select(entity_model).selected_columns["id"]

    if show_elements is None:
        show_elements = []

    extra_cols = []

    # collect the column objects for the select list
    for mi in show_elements:
        if mi not in entity_columns:
            raise ValueError(f"{mi} is not a valid column for {entity_type}")
        else:
            extra_cols.append(select(entity_model).selected_columns[mi])

    if name_like is not None:
        if "name" not in entity_columns:
            raise (ValueError("To filter by name requires name in the show_elements list."))

    attr = db.get_view("eattributes")
    # id_col = attr.c.entity.label("id")
    cols = [entity_id_col]
    cols.extend(extra_cols)
    more_info_cols = cols.copy()
    cols.extend(
        [
            attr.c.attr_id.label(atr_id_column_name),
            attr.c.the_type.label(type_column_name),
            attr.c.the_value.label(column_name),
            attr.c.the_date.label(date_column_name),
            attr.c.the_line.label(line_column_name),
            attr.c.the_level.label(level_column_name),
            attr.c.aobs.label(obs_column_name),
            attr.c.a_extra_info.label(attribute_extra_info_column_name),
        ]
    )

    if type(the_type) is list:
        attribute_query = attr.c.the_type.in_(the_type)
    else:
        attribute_query = attr.c.the_type.like(the_type)

    # filter by id list
    if filter_by is not None:
        # in some cases the filter_by list
        #  may contain ids that are not in the attribute table
        #  we need to add them to the final dataframe
        filter_by_sql = (
            select(entity_model)
            .with_only_columns(*more_info_cols, maintain_column_froms=True)
            .where(entity_model.id.in_(filter_by))
        )
        with db.session() as session:
            filtered_by_rows = session.execute(filter_by_sql)
            col_names = filter_by_sql.selected_columns.keys()
            filtered_df = pd.DataFrame.from_records(
                filtered_by_rows, index=["id"], columns=col_names
            )

        stmt = (
            (select(entity_model).where(entity_id_col.in_(filter_by)))
            .join(attr, attr.c.entity == entity_model.id, isouter=True)
            .where(attribute_query)
            .with_only_columns(*cols)
        )
    else:
        stmt = (
            select(entity_model)
            .join(attr, attr.c.entity == entity_model.id)
            .where(attribute_query)
            .with_only_columns(*cols)
        )

    # Filter by value
    if the_value is not None:
        if type(the_value) is list:
            stmt = stmt.where(attr.c.the_value.in_(the_value))
        elif type(the_value) is str:
            stmt = stmt.where(attr.c.the_value.like(the_value))
        else:
            raise ValueError("the_value must be either a string or a list of strings")

    # filter by date
    if dates_in is not None:
        after_date, before_date = dates_in

        stmt = stmt.where(attr.c.the_date > after_date, attr.c.the_date < before_date)

    # filter by name
    if name_like is not None:
        stmt = stmt.where(select(entity_model).selected_columns["name"].like(name_like))

    stmt = stmt.order_by(attr.c.the_date)

    if sql_echo:
        print(f"Query for {the_type}:\n", stmt)

    with db.session() as session:
        records = session.execute(stmt)
        col_names = stmt.selected_columns.keys()
        df = pd.DataFrame.from_records(records, index=["id"], columns=col_names)
        if df.iloc[0].count() == 0:
            return None  # nothing found we return None
        # Check for extra info
        extra_info_edits = []
        for row_number, (_, row) in enumerate(df.iterrows()):
            # Perform operations using row_number, index, and row
            if row[attribute_extra_info_column_name] is not None:
                # get the extra_info dict from the row
                # the dict contains information store during the import
                # process that is not stored directly in the attribute/columns
                # of the database. These include: comment and original aspects,
                # the element name and class in the original source
                # and the attribute name and column name in the database
                extra_info: dict = row[attribute_extra_info_column_name]
                for key, value in extra_info.items():
                    if key != "the_value":
                        # here we determine the name of columns
                        # with extra information about the value of
                        # the kleio attribute (group attribute or descendatns)
                        # normally this would be comments or original wording
                        # of the_type or the_date.
                        # we need to use the name of the ORM attribute
                        # "type" or "date" instead of the column name
                        # and this is stored in "entity_attr_name" in
                        # the extra_info dict under the key for the column name.
                        attr_name = value.get("entity_attr_name", key)
                        xtra_col_name = f"{column_name}.{attr_name}"
                    else:
                        xtra_col_name = column_name
                    original = value.get("original", None)
                    if original is not None:
                        xtra_col_name = f"{xtra_col_name}.original"
                        extra_info_edits.append((row_number, xtra_col_name, original))
                    comment = value.get("comment", None)
                    if comment is not None:
                        xtra_col_name = f"{xtra_col_name}.comment"
                        extra_info_edits.append((row_number, xtra_col_name, comment))
        for row_number, xtra_col_name, avalue in extra_info_edits:
            if xtra_col_name not in df.columns:
                df[xtra_col_name] = None
            df.iat[row_number, df.columns.get_loc(xtra_col_name)] = avalue

    if filter_by is not None:
        fb_ids = filtered_df.index.unique()
        attr_ids = df.index.unique()
        missing = list(set(fb_ids) - set(attr_ids))
        if len(missing) > 0:
            missing_df = filtered_df.loc[missing]
            df = pd.concat([df, missing_df])

    if df.iloc[0].count() == 0:
        return None  # nothing found we

    if more_attributes is None:
        more_columns = []
    else:
        more_columns = more_attributes

    if len(more_columns) > 0:
        for mcol in more_columns:
            column_name = mcol
            date_column_name = f"{column_name}.date"
            obs_column_name = f"{column_name}.obs"
            extra_info_column_name = f"{column_name}.extra_info"
            stmt = select(
                attr.c.entity.label("id"),
                attr.c.the_value.label(column_name),
                attr.c.the_date.label(date_column_name),
                attr.c.aobs.label(obs_column_name),
                attr.c.a_extra_info.label(extra_info_column_name),
            ).where(attr.c.the_type == mcol
                    ).where(attr.c.entity.in_(df.index))
            # col_names = stmt.columns.keys()

            with db.session() as session:
                records = session.execute(stmt)
                col_names = stmt.selected_columns.keys()

                df2 = pd.DataFrame.from_records(
                    records, index=["id"], columns=col_names
                )
                if df2.iloc[0].count() > 0:
                    extra_info_edits = []
                    for row_number, (_, row) in enumerate(df2.iterrows()):
                        if row[extra_info_column_name] is not None:
                            extra_info: dict = row[extra_info_column_name]
                            for key, value in extra_info.items():
                                if key != "the_value":
                                    attr_name = value.get("entity_attr_name", key)
                                    xtra_col_name2 = f"{column_name}.{attr_name}"
                                else:
                                    xtra_col_name2 = column_name
                                original = value.get("original", None)
                                if original is not None:
                                    xtra_col_name2 = f"{xtra_col_name2}.original"
                                    extra_info_edits.append((row_number, xtra_col_name2, original))
                                comment = value.get("comment", None)
                                if comment is not None:
                                    xtra_col_name2 = f"{xtra_col_name2}.comment"
                    for row_number, xtra_col_name2, avalue in extra_info_edits:
                        if xtra_col_name2 not in df2.columns:
                            df2[xtra_col_name2] = None
                        df2.iat[row_number, df2.columns.get_loc(xtra_col_name2)] = avalue

            if sql_echo:
                print(f"Query for more_attributes={mcol}:\n", stmt)

            if df2.iloc[0].count() == 0:
                df[mcol] = None  # nothing found we set the column to nulls
            else:
                df = df.join(df2)

    return df



""" Module  for SQLAlchemy and Pydantic models of the Timelink Database.

TODO: after testing for MHK schema:
    1. Implemente TemporalEntity with flexible dates
    2. Check the dataclasses new mapping https://docs.sqlalchemy.org/en/20/orm/dataclasses.html#orm-declarative-native-dataclasses
"""
from .base_class import Base  # noqa pylint: disable=unused-import
from .entity import Entity  # noqa pylint: disable=unused-import
from .attribute import Attribute  # noqa pylint: disable=unused-import
from .relation import Relation  # noqa pylint: disable=unused-import
from .act import Act  # noqa pylint: disable=unused-import
from .source import Source  # noqa pylint: disable=unused-import
from .aregister import ARegister  # noqa pylint: disable=unused-import
from .person import Person  # noqa pylint: disable=unused-import
from .object import Object  # noqa pylint: disable=unused-import
from .geoentity import Geoentity  # noqa pylint: disable=unused-import
from .rentity import REntity, Link  # noqa pylint: disable=unused-import
from .pom_som_mapper import PomSomMapper  # noqa pylint: disable=unused-import
from .pom_som_mapper import PomClassAttributes  # noqa pylint: disable=unused-import
from .base_mappings import pom_som_base_mappings  # noqa pylint: disable=unused-import
from .system import SysPar  # noqa pylint: disable=unused-import
from .system import SysParSchema  # noqa pylint: disable=unused-import
from .system import SysLog  # noqa pylint: disable=unused-import
from .system import SysLogSchema  # noqa pylint: disable=unused-import
from .system import SysLogCreateSchema  # noqa pylint: disable=unused-import
from .system import KleioImportedFile  # noqa pylint: disable=unused-import



from typing import Optional

# pylint: disable=import-error
from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import mapped_column, Mapped

from timelink.kleio.utilities import quote_long_text
from .entity import Entity


class Person(Entity):
    """Represents a person in a historical source"""

    __tablename__ = "persons"

    id: Mapped[str] = mapped_column(
        String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True
    )
    name: Mapped[Optional[str]] = mapped_column(String, index=True)
    sex: Mapped[Optional[str]] = mapped_column(String(1))
    obs: Mapped[Optional[str]] = mapped_column(String)

    __mapper_args__ = {"polymorphic_identity": "person"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Person(id={sr}, "
            f'name="{self.name}", '
            f'sex="{self.sex}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__(self):
        return self.to_kleio()

    def to_kleio(self, self_string='', ident="", ident_inc="  ", show_contained=True, width=80, **kwargs) -> str:
        if self.groupname is None:
            myname = "person"
        else:
            myname = self.groupname
        r = (
            f"{myname}${self.for_kleio('name')}/"
            f"{self.for_kleio('sex')}{self.render_id()}"
        )
        obss = self.for_kleio('obs')
        if len(obss) > 0:
            r = f"{r}/obs={quote_long_text(obss,width=width)}"
        kleio = super().to_kleio(
            self_string=r,
            show_contained=show_contained,
            ident=ident,
            ident_inc=ident_inc,
            width=width,
            **kwargs
        )
        return kleio


def get_person(id: str = None, db=None, session=None, sql_echo: bool = False) -> Person:
    """
    Fetch a person from the database
    """
    if id is None:
        raise ValueError("Error, id needed")

    if session is None and db is None:
        raise ValueError("Error, session or database object needed")

    if db is not None:
        with db.session() as session:
            p: Person = session.get(Person, id)
    elif session is not None:
        p: Person = session.get(Person, id)
    return p


def pperson(id: str, session=None):
    """Prints a person in kleio notation"""
    print(get_person(id=id, session=session).to_kleio())



"""Add source column to Entity

Revision ID: a6d2e17ecfb1
Revises:
Create Date: 2024-10-27 18:43:22.039866

"""
from typing import Sequence, Union

from alembic import op
from sqlalchemy import Column, String

from timelink.migrations import column_exists


# revision identifiers, used by Alembic.
revision: str = 'a6d2e17ecfb1'
down_revision: Union[str, None] = "89f407c298c8"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    if not column_exists('entities', 'the_source', op):
        op.add_column('entities', Column('the_source', String, nullable=True))


def downgrade() -> None:
    if column_exists('entities', 'the_source', op):
        op.drop_column('entities', 'the_source')



"""
(c) Joaquim Carvalho 2021.
MIT License, no warranties.
"""

from typing import Optional
from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import mapped_column, Mapped

from timelink.kleio.utilities import quote_long_text
from .entity import Entity


class Object(Entity):
    """
    Represent an object in the database. Objects are entities that are not people.
    """
    __tablename__ = "objects"

    id: Mapped[str] = mapped_column(String, ForeignKey("entities.id", ondelete="CASCADE"), primary_key=True)
    name: Mapped[Optional[str]] = mapped_column(String, index=True)
    the_type: Mapped[str] = mapped_column(String(32), index=True)
    obs: Mapped[Optional[str]] = mapped_column(String)

    __mapper_args__ = {"polymorphic_identity": "object"}

    def __repr__(self):
        sr = super().__repr__()
        return (
            f"Object(id={sr}, "
            f'name="{self.name}", '
            f'the_type="{self.the_type}", '
            f"obs={self.obs}"
            f")"
        )

    def __str__XX(self):
        return self.to_kleio(show_contained=False)

    def to_kleio_xx(self, ident="", ident_inc="  ", width=80, show_contained=True, **kwargs) -> str:
        if self.name is None:
            name = ""
        else:
            name = self.name + "/"
        r = f"{self.groupname}${name}{quote_long_text(self.the_type, width=width)}{self.render_id()}"
        if self.obs is not None and len(self.obs.strip()) > 0:
            r = f"{r}/obs={quote_long_text(self.obs.strip(), width=width)}"
        r = super().to_kleio(self_string=r,
                             ident=ident,
                             ident_inc=ident_inc,
                             width=width,
                             show_contained=show_contained)
        return r
